/**
 *                     Sidplayer and Stereo Sidplayer format support.
 *                     ----------------------------------------------
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package libsidplay.sidtune;

import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Arrays;

class Mus extends PSid {

	//
	// sidtune format errors
	//

	private static final String _sidtune_txt_invalid = "ERROR: File contains invalid data";

	private static final String _sidtune_2nd_invalid = "ERROR: 2nd file contains invalid data";

	private static final int MUS_HLT_CMD = 0x14F;

	private static final int MUS_DATA_ADDR = 0x0900;

	private static final int STR_SID2_ADDR = 0xd500;

	/**
	 * Needed for MUS/STR player installation.
	 */
	private int musDataLen;

	@Override
	public int placeProgramInMemory(final byte[] c64buf) {
		installPlayer(c64buf);
		return super.placeProgramInMemory(c64buf);
	}

	@Override
	public void save(final String destFileName, final boolean overWriteFlag) throws IOException {
		final FileOutputStream fMyOut = new FileOutputStream(destFileName, !overWriteFlag);
		fMyOut.write(program);
		fMyOut.close();
	}

	private static boolean detect(final byte[] buffer, final int startIndex, final int[] voice3Index) {
		if (buffer == null || buffer.length < startIndex + 8) {
			return false;
		}

		// Add length of voice 1 data.
		final int voice1Index = 2 + 3 * 2 + ((buffer[startIndex + 2] & 0xff) + ((buffer[startIndex + 3] & 0xff) << 8));
		// Add length of voice 2 data.
		final int voice2Index = voice1Index + ((buffer[startIndex + 4] & 0xff) + ((buffer[startIndex + 5] & 0xff) << 8));
		// Add length of voice 3 data.
		voice3Index[0] = voice2Index + ((buffer[startIndex + 6] & 0xff) + ((buffer[startIndex + 7] & 0xff) << 8));

		/* validate that voice3 is still inside the buffer and that each track ends with HLT */
		return voice3Index[0] - 1 < buffer.length
		&& ((buffer[voice1Index-1] & 0xff) + ((buffer[voice1Index-2] & 0xff) << 8)) == MUS_HLT_CMD
		&& ((buffer[voice2Index-1] & 0xff) + ((buffer[voice2Index-2] & 0xff) << 8)) == MUS_HLT_CMD
		&& ((buffer[voice3Index[0]-1] & 0xff) + ((buffer[voice3Index[0]-2] & 0xff) << 8)) == MUS_HLT_CMD;
	}

	public static SidTune load(final byte[] musBuf, final byte[] strBuf) throws SidTuneError {
		final Mus sidtune = new Mus();
		sidtune.info.compatibility = Compatibility.PSIDv2;
		return loadWithProvidedMetadata(musBuf, strBuf, sidtune);
	}

	public static SidTune loadWithProvidedMetadata(final byte[] musBuf, byte[] strBuf, final Mus sidtune) throws SidTuneError {
		final int[] voice3Index = new int[1];
		if (!detect(musBuf, sidtune.fileOffset, voice3Index)) {
			return null;
		}

		sidtune.info.songs = sidtune.info.startSong = 1;
		sidtune.convertOldStyleSpeedToTables(~0);

		// Check setting compatibility for MUS playback
		if (sidtune.info.compatibility != Compatibility.PSIDv2
				|| sidtune.info.relocStartPage != 0 || sidtune.info.relocPages != 0) {
			throw new SidTuneError(_sidtune_txt_invalid);
		}

		for (int i = 0; i < sidtune.info.songs; i++) {
			if (sidtune.songSpeed[i] != Speed.CIA_1A) {
				throw new SidTuneError(_sidtune_txt_invalid);
			}
		}

		sidtune.musDataLen = musBuf.length;
		sidtune.info.loadAddr = MUS_DATA_ADDR;

		int infoStringLocation = voice3Index[0];

		/* FIXME: dealing with credits is problematic. The data is unstructured, and longer
		 * than fits into DB index. We could assume that first line is title, second is author,
		 * third is release, but that's actually not very likely. */
		{
			final String credit = convertPetsciiToAscii(musBuf, infoStringLocation);
			infoStringLocation += credit.length() + 1;
			sidtune.info.commentString[0] = credit;
			sidtune.info.numberOfCommentStrings = 1;
		}

		// If we appear to have additional data at the end, check is it's
		// another mus file (but only if a second file isn't supplied)
		if (strBuf == null && infoStringLocation < musBuf.length) {
			strBuf = Arrays.copyOfRange(musBuf, infoStringLocation, musBuf.length);
		}

		if (strBuf != null) {
			if (!detect(strBuf, 0, voice3Index)) {
				throw new RuntimeException(_sidtune_2nd_invalid);
			}

			infoStringLocation = voice3Index[0];

			{
				final String credit = SidTune.convertPetsciiToAscii(strBuf, infoStringLocation);
				infoStringLocation += credit.length() + 1;
				sidtune.info.commentString[1] = credit;
				sidtune.info.numberOfCommentStrings = 2;
			}

			sidtune.info.sidChipBase2 = STR_SID2_ADDR;
			sidtune.info.initAddr = 0xfc90;
			sidtune.info.playAddr = 0xfc96;
		} else {
			sidtune.info.initAddr = 0xec60;
			sidtune.info.playAddr = 0xec80;
		}

		if (sidtune.info.infoString[0] == null) {
			sidtune.info.infoString[0] = "<?>";
		}
		if (sidtune.info.infoString[1] == null) {
			sidtune.info.infoString[1] = "<?>";
		}
		if (sidtune.info.infoString[2] == null) {
			sidtune.info.infoString[2] = "<?>";
		}

		sidtune.program = new byte[musBuf.length + (strBuf != null ? strBuf.length : 0)];
		sidtune.info.c64dataLen = sidtune.program.length - sidtune.fileOffset;
		System.arraycopy(musBuf, 0, sidtune.program, 0, musBuf.length);
		if (strBuf != null) {
			System.arraycopy(strBuf, 0, sidtune.program, musBuf.length, strBuf.length);
		}

		sidtune.findPlaceForDriver();

		return sidtune;
	}

	private static final byte _sidtune_sidplayer1[] = {

		(byte) 0x00, (byte) 0xE0, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x61, (byte) 0xE1, (byte) 0x60, (byte) 0x01, (byte) 0x02, (byte) 0x04, (byte) 0x00, 
		(byte) 0x07, (byte) 0x0E, (byte) 0x02, (byte) 0x02, (byte) 0xFE, (byte) 0x02, (byte) 0x02, (byte) 0xFE, 
		(byte) 0xFE, (byte) 0x00, (byte) 0x01, (byte) 0x00, (byte) 0xFF, (byte) 0x00, (byte) 0x02, (byte) 0x04, 
		(byte) 0x05, (byte) 0x07, (byte) 0x09, (byte) 0x0B, (byte) 0x1E, (byte) 0x18, (byte) 0x8B, (byte) 0x7E, 
		(byte) 0xFA, (byte) 0x06, (byte) 0xAC, (byte) 0xF3, (byte) 0xE6, (byte) 0x8F, (byte) 0xF8, (byte) 0x2E, 
		(byte) 0x86, (byte) 0x8E, (byte) 0x96, (byte) 0x9F, (byte) 0xA8, (byte) 0xB3, (byte) 0xBD, (byte) 0xC8, 
		(byte) 0xD4, (byte) 0xE1, (byte) 0xEE, (byte) 0xFD, (byte) 0x8C, (byte) 0x78, (byte) 0x64, (byte) 0x50, 
		(byte) 0x3C, (byte) 0x28, (byte) 0x14, (byte) 0x00, (byte) 0x00, (byte) 0x02, (byte) 0x03, (byte) 0x05, 
		(byte) 0x07, (byte) 0x08, (byte) 0x0A, (byte) 0x0C, (byte) 0x0D, (byte) 0x0F, (byte) 0x11, (byte) 0x12, 
		(byte) 0x00, (byte) 0xE0, (byte) 0x00, (byte) 0x05, (byte) 0x0A, (byte) 0x0F, (byte) 0xF9, (byte) 0x00, 
		(byte) 0xF5, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x10, (byte) 0x00, (byte) 0x00, (byte) 0x20, 
		(byte) 0x00, (byte) 0x00, (byte) 0x30, (byte) 0x00, (byte) 0x00, (byte) 0x40, (byte) 0x00, (byte) 0x00, 
		(byte) 0x50, (byte) 0x00, (byte) 0x00, (byte) 0x60, (byte) 0x00, (byte) 0x00, (byte) 0x70, (byte) 0x00, 
		(byte) 0x00, (byte) 0x80, (byte) 0x00, (byte) 0x00, (byte) 0x90, (byte) 0x00, (byte) 0x00, (byte) 0xA0, 
		(byte) 0x00, (byte) 0xA9, (byte) 0x00, (byte) 0x8D, (byte) 0x00, (byte) 0xE0, (byte) 0xA2, (byte) 0x95, 
		(byte) 0xA0, (byte) 0x42, (byte) 0xAD, (byte) 0xA6, (byte) 0x02, (byte) 0xF0, (byte) 0x04, (byte) 0xA2, 
		(byte) 0x25, (byte) 0xA0, (byte) 0x40, (byte) 0x8E, (byte) 0x5B, (byte) 0xE1, (byte) 0x8C, (byte) 0x5C, 
		(byte) 0xE1, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, 
		(byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, 
		(byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, 
		(byte) 0xEA, (byte) 0x60, (byte) 0xA9, (byte) 0x00, (byte) 0x8D, (byte) 0x00, (byte) 0xE0, (byte) 0x86, 
		(byte) 0x61, (byte) 0x84, (byte) 0x62, (byte) 0xA0, (byte) 0xBC, (byte) 0x99, (byte) 0x00, (byte) 0xE0, 
		(byte) 0x88, (byte) 0xD0, (byte) 0xFA, (byte) 0xA0, (byte) 0x72, (byte) 0x99, (byte) 0xBC, (byte) 0xE0, 
		(byte) 0x88, (byte) 0xD0, (byte) 0xFA, (byte) 0x8D, (byte) 0x15, (byte) 0xD4, (byte) 0x8D, (byte) 0x16, 
		(byte) 0xD4, (byte) 0xA9, (byte) 0x08, (byte) 0x8D, (byte) 0x25, (byte) 0xE0, (byte) 0x8D, (byte) 0x17, 
		(byte) 0xD4, (byte) 0x8D, (byte) 0x26, (byte) 0xE0, (byte) 0x8D, (byte) 0x18, (byte) 0xD4, (byte) 0xA9, 
		(byte) 0x90, (byte) 0x8D, (byte) 0x27, (byte) 0xE0, (byte) 0xA9, (byte) 0x60, (byte) 0x8D, (byte) 0x28, 
		(byte) 0xE0, (byte) 0xA9, (byte) 0x0C, (byte) 0x8D, (byte) 0x29, (byte) 0xE0, (byte) 0xAD, (byte) 0x5B, 
		(byte) 0xE1, (byte) 0x8D, (byte) 0x2D, (byte) 0xE0, (byte) 0xAD, (byte) 0x5C, (byte) 0xE1, (byte) 0x8D, 
		(byte) 0x2E, (byte) 0xE0, (byte) 0xA9, (byte) 0xFF, (byte) 0x8D, (byte) 0xCC, (byte) 0xE0, (byte) 0xA9, 
		(byte) 0xD4, (byte) 0x85, (byte) 0x64, (byte) 0xA2, (byte) 0x02, (byte) 0xA9, (byte) 0xFF, (byte) 0x9D, 
		(byte) 0x0B, (byte) 0xE0, (byte) 0xA9, (byte) 0x01, (byte) 0x9D, (byte) 0x30, (byte) 0xE0, (byte) 0x9D, 
		(byte) 0x2A, (byte) 0xE0, (byte) 0x8A, (byte) 0x9D, (byte) 0x33, (byte) 0xE0, (byte) 0x9D, (byte) 0xAE, 
		(byte) 0xE0, (byte) 0xA9, (byte) 0x04, (byte) 0x9D, (byte) 0x39, (byte) 0xE0, (byte) 0xBD, (byte) 0xA8, 
		(byte) 0xE1, (byte) 0x9D, (byte) 0xBA, (byte) 0xE0, (byte) 0xA9, (byte) 0x5B, (byte) 0x9D, (byte) 0x7E, 
		(byte) 0xE0, (byte) 0xBD, (byte) 0x65, (byte) 0xE1, (byte) 0x85, (byte) 0x63, (byte) 0xA9, (byte) 0x00, 
		(byte) 0xA8, (byte) 0x91, (byte) 0x63, (byte) 0xC8, (byte) 0x91, (byte) 0x63, (byte) 0xC8, (byte) 0x91, 
		(byte) 0x63, (byte) 0xA9, (byte) 0x08, (byte) 0x9D, (byte) 0x17, (byte) 0xE0, (byte) 0x9D, (byte) 0x9C, 
		(byte) 0xE0, (byte) 0xC8, (byte) 0x91, (byte) 0x63, (byte) 0xC8, (byte) 0x91, (byte) 0x63, (byte) 0xA9, 
		(byte) 0x40, (byte) 0x9D, (byte) 0x1A, (byte) 0xE0, (byte) 0x91, (byte) 0x63, (byte) 0xA9, (byte) 0x20, 
		(byte) 0x9D, (byte) 0x1D, (byte) 0xE0, (byte) 0xC8, (byte) 0x91, (byte) 0x63, (byte) 0xA9, (byte) 0xF5, 
		(byte) 0x9D, (byte) 0x20, (byte) 0xE0, (byte) 0xC8, (byte) 0x91, (byte) 0x63, (byte) 0xCA, (byte) 0x10, 
		(byte) 0xA4, (byte) 0x8A, (byte) 0xA2, (byte) 0x17, (byte) 0x9D, (byte) 0x3E, (byte) 0xE1, (byte) 0xCA, 
		(byte) 0x10, (byte) 0xFA, (byte) 0xA5, (byte) 0x61, (byte) 0x18, (byte) 0x69, (byte) 0x06, (byte) 0x85, 
		(byte) 0x63, (byte) 0xA9, (byte) 0x00, (byte) 0xAA, (byte) 0xA8, (byte) 0x65, (byte) 0x62, (byte) 0x85, 
		(byte) 0x64, (byte) 0x9D, (byte) 0xAB, (byte) 0xE0, (byte) 0x9D, (byte) 0xB4, (byte) 0xE0, (byte) 0xA5, 
		(byte) 0x63, (byte) 0x9D, (byte) 0xA8, (byte) 0xE0, (byte) 0x9D, (byte) 0xB1, (byte) 0xE0, (byte) 0x18, 
		(byte) 0x71, (byte) 0x61, (byte) 0x85, (byte) 0x63, (byte) 0xA5, (byte) 0x64, (byte) 0xC8, (byte) 0x71, 
		(byte) 0x61, (byte) 0xC8, (byte) 0xE8, (byte) 0xE0, (byte) 0x03, (byte) 0xD0, (byte) 0xE0, (byte) 0xA6, 
		(byte) 0x63, (byte) 0xA8, (byte) 0x60, (byte) 0xA9, (byte) 0x00, (byte) 0x8D, (byte) 0x04, (byte) 0xD4, 
		(byte) 0x8D, (byte) 0x0B, (byte) 0xD4, (byte) 0x8D, (byte) 0x12, (byte) 0xD4, (byte) 0x8D, (byte) 0x01, 
		(byte) 0xD4, (byte) 0x8D, (byte) 0x08, (byte) 0xD4, (byte) 0x8D, (byte) 0x0F, (byte) 0xD4, (byte) 0xA9, 
		(byte) 0x08, (byte) 0x8D, (byte) 0x17, (byte) 0xD4, (byte) 0xAD, (byte) 0x5B, (byte) 0xE1, (byte) 0x8D, 
		(byte) 0x04, (byte) 0xDC, (byte) 0xAD, (byte) 0x5C, (byte) 0xE1, (byte) 0x8D, (byte) 0x05, (byte) 0xDC, 
		(byte) 0x60, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, 
		(byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0x60, 
		(byte) 0xA9, (byte) 0x08, (byte) 0x8D, (byte) 0x00, (byte) 0xE0, (byte) 0x6C, (byte) 0x5D, (byte) 0xE1, 
		(byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xAD, (byte) 0x00, (byte) 0xE0, (byte) 0x30, (byte) 0xF0, 
		(byte) 0x09, (byte) 0x80, (byte) 0xA8, (byte) 0x29, (byte) 0x07, (byte) 0xF0, (byte) 0xEE, (byte) 0xD8, 
		(byte) 0x8C, (byte) 0x00, (byte) 0xE0, (byte) 0xEA, (byte) 0xA5, (byte) 0xFB, (byte) 0x8D, (byte) 0x56, 
		(byte) 0xE1, (byte) 0xA5, (byte) 0xFC, (byte) 0x8D, (byte) 0x57, (byte) 0xE1, (byte) 0xA5, (byte) 0xFD, 
		(byte) 0x8D, (byte) 0x58, (byte) 0xE1, (byte) 0xA5, (byte) 0xFE, (byte) 0x8D, (byte) 0x59, (byte) 0xE1, 
		(byte) 0xA5, (byte) 0xFF, (byte) 0x8D, (byte) 0x5A, (byte) 0xE1, (byte) 0xAD, (byte) 0x23, (byte) 0xE0, 
		(byte) 0x18, (byte) 0x6D, (byte) 0xD9, (byte) 0xE0, (byte) 0x48, (byte) 0x29, (byte) 0x07, (byte) 0xA8, 
		(byte) 0xAD, (byte) 0xDC, (byte) 0xE0, (byte) 0x69, (byte) 0x00, (byte) 0x85, (byte) 0xFF, (byte) 0x68, 
		(byte) 0x46, (byte) 0xFF, (byte) 0x6A, (byte) 0x46, (byte) 0xFF, (byte) 0x6A, (byte) 0x46, (byte) 0xFF, 
		(byte) 0x6A, (byte) 0x18, (byte) 0x6D, (byte) 0x24, (byte) 0xE0, (byte) 0x8C, (byte) 0x15, (byte) 0xD4, 
		(byte) 0x8D, (byte) 0x16, (byte) 0xD4, (byte) 0xAD, (byte) 0x25, (byte) 0xE0, (byte) 0x8D, (byte) 0x17, 
		(byte) 0xD4, (byte) 0xAD, (byte) 0x26, (byte) 0xE0, (byte) 0x8D, (byte) 0x18, (byte) 0xD4, (byte) 0xA9, 
		(byte) 0xD4, (byte) 0x85, (byte) 0xFC, (byte) 0xA2, (byte) 0x00, (byte) 0xAD, (byte) 0x00, (byte) 0xE0, 
		(byte) 0x3D, (byte) 0x62, (byte) 0xE1, (byte) 0xF0, (byte) 0x51, (byte) 0xBD, (byte) 0x65, (byte) 0xE1, 
		(byte) 0x85, (byte) 0xFB, (byte) 0xBD, (byte) 0x0E, (byte) 0xE0, (byte) 0x18, (byte) 0x7D, (byte) 0x51, 
		(byte) 0xE0, (byte) 0xA8, (byte) 0xBD, (byte) 0x11, (byte) 0xE0, (byte) 0x7D, (byte) 0x54, (byte) 0xE0, 
		(byte) 0x48, (byte) 0x98, (byte) 0x18, (byte) 0x7D, (byte) 0xCD, (byte) 0xE0, (byte) 0xA0, (byte) 0x00, 
		(byte) 0x91, (byte) 0xFB, (byte) 0x68, (byte) 0x7D, (byte) 0xD0, (byte) 0xE0, (byte) 0xC8, (byte) 0x91, 
		(byte) 0xFB, (byte) 0xBD, (byte) 0x14, (byte) 0xE0, (byte) 0x18, (byte) 0x7D, (byte) 0x69, (byte) 0xE0, 
		(byte) 0x85, (byte) 0xFF, (byte) 0xBD, (byte) 0x17, (byte) 0xE0, (byte) 0x7D, (byte) 0x6C, (byte) 0xE0, 
		(byte) 0x48, (byte) 0xA5, (byte) 0xFF, (byte) 0x18, (byte) 0x7D, (byte) 0xD3, (byte) 0xE0, (byte) 0xC8, 
		(byte) 0x91, (byte) 0xFB, (byte) 0x68, (byte) 0x7D, (byte) 0xD6, (byte) 0xE0, (byte) 0xC8, (byte) 0x91, 
		(byte) 0xFB, (byte) 0xBD, (byte) 0x1D, (byte) 0xE0, (byte) 0xC8, (byte) 0xC8, (byte) 0x91, (byte) 0xFB, 
		(byte) 0xBD, (byte) 0x20, (byte) 0xE0, (byte) 0xC8, (byte) 0x91, (byte) 0xFB, (byte) 0xE8, (byte) 0xE0, 
		(byte) 0x03, (byte) 0xD0, (byte) 0xA2, (byte) 0xAC, (byte) 0x1A, (byte) 0xE0, (byte) 0xAE, (byte) 0x1B, 
		(byte) 0xE0, (byte) 0xAD, (byte) 0x1C, (byte) 0xE0, (byte) 0x8C, (byte) 0x04, (byte) 0xD4, (byte) 0x8E, 
		(byte) 0x0B, (byte) 0xD4, (byte) 0x8D, (byte) 0x12, (byte) 0xD4, (byte) 0xAE, (byte) 0x2D, (byte) 0xE0, 
		(byte) 0xAC, (byte) 0x2E, (byte) 0xE0, (byte) 0x8E, (byte) 0x04, (byte) 0xDC, (byte) 0x8C, (byte) 0x05, 
		(byte) 0xDC, (byte) 0xAD, (byte) 0x1B, (byte) 0xD4, (byte) 0x8D, (byte) 0xBE, (byte) 0xE0, (byte) 0xAD, 
		(byte) 0x1C, (byte) 0xD4, (byte) 0x8D, (byte) 0xBF, (byte) 0xE0, (byte) 0xA2, (byte) 0x00, (byte) 0xAD, 
		(byte) 0x00, (byte) 0xE0, (byte) 0x3D, (byte) 0x62, (byte) 0xE1, (byte) 0xF0, (byte) 0x10, (byte) 0x8E, 
		(byte) 0x2F, (byte) 0xE0, (byte) 0x20, (byte) 0x36, (byte) 0xE5, (byte) 0xAD, (byte) 0x00, (byte) 0xE0, 
		(byte) 0x29, (byte) 0x78, (byte) 0xF0, (byte) 0x03, (byte) 0x4C, (byte) 0x0C, (byte) 0xE5, (byte) 0xE8, 
		(byte) 0xE0, (byte) 0x03, (byte) 0xD0, (byte) 0xE3, (byte) 0xAD, (byte) 0xC9, (byte) 0xE0, (byte) 0xD0, 
		(byte) 0x52, (byte) 0xAD, (byte) 0xCA, (byte) 0xE0, (byte) 0x0D, (byte) 0xCB, (byte) 0xE0, (byte) 0xF0, 
		(byte) 0x78, (byte) 0xAD, (byte) 0xDF, (byte) 0xE0, (byte) 0xD0, (byte) 0x28, (byte) 0xAD, (byte) 0xCA, 
		(byte) 0xE0, (byte) 0xF0, (byte) 0x28, (byte) 0x18, (byte) 0x6D, (byte) 0xBD, (byte) 0xE0, (byte) 0xB0, 
		(byte) 0x07, (byte) 0xCD, (byte) 0xCC, (byte) 0xE0, (byte) 0x90, (byte) 0x60, (byte) 0xF0, (byte) 0x5E, 
		(byte) 0xA9, (byte) 0x00, (byte) 0x8D, (byte) 0xDF, (byte) 0xE0, (byte) 0xAD, (byte) 0xCB, (byte) 0xE0, 
		(byte) 0xF0, (byte) 0x54, (byte) 0xEE, (byte) 0xDF, (byte) 0xE0, (byte) 0xAD, (byte) 0xBD, (byte) 0xE0, 
		(byte) 0xED, (byte) 0xCB, (byte) 0xE0, (byte) 0x4C, (byte) 0xB4, (byte) 0xE4, (byte) 0xAD, (byte) 0xCB, 
		(byte) 0xE0, (byte) 0xF0, (byte) 0xD3, (byte) 0xAD, (byte) 0xBD, (byte) 0xE0, (byte) 0x38, (byte) 0xED, 
		(byte) 0xCB, (byte) 0xE0, (byte) 0xB0, (byte) 0x3A, (byte) 0xA9, (byte) 0x00, (byte) 0x8D, (byte) 0xDF, 
		(byte) 0xE0, (byte) 0xAD, (byte) 0xCA, (byte) 0xE0, (byte) 0xD0, (byte) 0x30, (byte) 0xEE, (byte) 0xDF, 
		(byte) 0xE0, (byte) 0xD0, (byte) 0x28, (byte) 0xCE, (byte) 0xE0, (byte) 0xE0, (byte) 0xD0, (byte) 0x29, 
		(byte) 0xAD, (byte) 0xDF, (byte) 0xE0, (byte) 0xD0, (byte) 0x11, (byte) 0xEE, (byte) 0xDF, (byte) 0xE0, 
		(byte) 0xAD, (byte) 0xCB, (byte) 0xE0, (byte) 0xD0, (byte) 0x02, (byte) 0xA9, (byte) 0x20, (byte) 0x8D, 
		(byte) 0xE0, (byte) 0xE0, (byte) 0xA9, (byte) 0x00, (byte) 0xF0, (byte) 0x10, (byte) 0xCE, (byte) 0xDF, 
		(byte) 0xE0, (byte) 0xAD, (byte) 0xCA, (byte) 0xE0, (byte) 0xD0, (byte) 0x02, (byte) 0xA9, (byte) 0x20, 
		(byte) 0x8D, (byte) 0xE0, (byte) 0xE0, (byte) 0xAD, (byte) 0xCC, (byte) 0xE0, (byte) 0x8D, (byte) 0xBD, 
		(byte) 0xE0, (byte) 0xA2, (byte) 0x00, (byte) 0xBD, (byte) 0xC3, (byte) 0xE0, (byte) 0xF0, (byte) 0x44, 
		(byte) 0xA9, (byte) 0x00, (byte) 0x85, (byte) 0xFF, (byte) 0xBC, (byte) 0xC0, (byte) 0xE0, (byte) 0xB9, 
		(byte) 0xBD, (byte) 0xE0, (byte) 0xBC, (byte) 0xC6, (byte) 0xE0, (byte) 0xF0, (byte) 0x0E, (byte) 0x30, 
		(byte) 0x08, (byte) 0x0A, (byte) 0x26, (byte) 0xFF, (byte) 0x88, (byte) 0xD0, (byte) 0xFA, (byte) 0xF0, 
		(byte) 0x04, (byte) 0x4A, (byte) 0xC8, (byte) 0xD0, (byte) 0xFC, (byte) 0xBC, (byte) 0xC3, (byte) 0xE0, 
		(byte) 0x88, (byte) 0xD0, (byte) 0x0B, (byte) 0x9D, (byte) 0xCD, (byte) 0xE0, (byte) 0xA5, (byte) 0xFF, 
		(byte) 0x9D, (byte) 0xD0, (byte) 0xE0, (byte) 0x4C, (byte) 0x02, (byte) 0xE5, (byte) 0x88, (byte) 0xD0, 
		(byte) 0x0B, (byte) 0x9D, (byte) 0xD3, (byte) 0xE0, (byte) 0xA5, (byte) 0xFF, (byte) 0x9D, (byte) 0xD6, 
		(byte) 0xE0, (byte) 0x4C, (byte) 0x02, (byte) 0xE5, (byte) 0x8D, (byte) 0xD9, (byte) 0xE0, (byte) 0xA5, 
		(byte) 0xFF, (byte) 0x8D, (byte) 0xDC, (byte) 0xE0, (byte) 0xE8, (byte) 0xE0, (byte) 0x03, (byte) 0xD0, 
		(byte) 0xB2, (byte) 0xAD, (byte) 0x00, (byte) 0xE0, (byte) 0x29, (byte) 0x7F, (byte) 0x8D, (byte) 0x00, 
		(byte) 0xE0, (byte) 0xAD, (byte) 0x56, (byte) 0xE1, (byte) 0x85, (byte) 0xFB, (byte) 0xAD, (byte) 0x57, 
		(byte) 0xE1, (byte) 0x85, (byte) 0xFC, (byte) 0xAD, (byte) 0x58, (byte) 0xE1, (byte) 0x85, (byte) 0xFD, 
		(byte) 0xAD, (byte) 0x59, (byte) 0xE1, (byte) 0x85, (byte) 0xFE, (byte) 0xAD, (byte) 0x5A, (byte) 0xE1, 
		(byte) 0x85, (byte) 0xFF, (byte) 0x6C, (byte) 0x5D, (byte) 0xE1, (byte) 0xBD, (byte) 0x60, (byte) 0xE0, 
		(byte) 0xD0, (byte) 0x03, (byte) 0x4C, (byte) 0x9F, (byte) 0xE6, (byte) 0x4C, (byte) 0xBA, (byte) 0xE5, 
		(byte) 0xDE, (byte) 0x30, (byte) 0xE0, (byte) 0xD0, (byte) 0x03, (byte) 0x4C, (byte) 0xA0, (byte) 0xE6, 
		(byte) 0xBD, (byte) 0x36, (byte) 0xE0, (byte) 0x30, (byte) 0xE8, (byte) 0xD0, (byte) 0x1A, (byte) 0xBD, 
		(byte) 0x3F, (byte) 0xE0, (byte) 0xF0, (byte) 0x05, (byte) 0xDE, (byte) 0x3F, (byte) 0xE0, (byte) 0xD0, 
		(byte) 0x10, (byte) 0xBD, (byte) 0x39, (byte) 0xE0, (byte) 0xDD, (byte) 0x30, (byte) 0xE0, (byte) 0x90, 
		(byte) 0x08, (byte) 0xBD, (byte) 0x1A, (byte) 0xE0, (byte) 0x29, (byte) 0xFE, (byte) 0x9D, (byte) 0x1A, 
		(byte) 0xE0, (byte) 0xBD, (byte) 0x42, (byte) 0xE0, (byte) 0xF0, (byte) 0x56, (byte) 0x0A, (byte) 0xBD, 
		(byte) 0x0E, (byte) 0xE0, (byte) 0xB0, (byte) 0x1D, (byte) 0x7D, (byte) 0x45, (byte) 0xE0, (byte) 0x9D, 
		(byte) 0x0E, (byte) 0xE0, (byte) 0xA8, (byte) 0xBD, (byte) 0x11, (byte) 0xE0, (byte) 0x7D, (byte) 0x48, 
		(byte) 0xE0, (byte) 0x9D, (byte) 0x11, (byte) 0xE0, (byte) 0x48, (byte) 0x98, (byte) 0xDD, (byte) 0x8D, 
		(byte) 0xE0, (byte) 0x68, (byte) 0xFD, (byte) 0x90, (byte) 0xE0, (byte) 0xB0, (byte) 0x1F, (byte) 0x90, 
		(byte) 0x2E, (byte) 0xFD, (byte) 0x45, (byte) 0xE0, (byte) 0x9D, (byte) 0x0E, (byte) 0xE0, (byte) 0xBD, 
		(byte) 0x11, (byte) 0xE0, (byte) 0xFD, (byte) 0x48, (byte) 0xE0, (byte) 0x9D, (byte) 0x11, (byte) 0xE0, 
		(byte) 0xBD, (byte) 0x8D, (byte) 0xE0, (byte) 0xDD, (byte) 0x0E, (byte) 0xE0, (byte) 0xBD, (byte) 0x90, 
		(byte) 0xE0, (byte) 0xFD, (byte) 0x11, (byte) 0xE0, (byte) 0x90, (byte) 0x11, (byte) 0xBD, (byte) 0x8D, 
		(byte) 0xE0, (byte) 0x9D, (byte) 0x0E, (byte) 0xE0, (byte) 0xBD, (byte) 0x90, (byte) 0xE0, (byte) 0x9D, 
		(byte) 0x11, (byte) 0xE0, (byte) 0xA9, (byte) 0x00, (byte) 0x9D, (byte) 0x42, (byte) 0xE0, (byte) 0xBD, 
		(byte) 0x60, (byte) 0xE0, (byte) 0xF0, (byte) 0x55, (byte) 0xBD, (byte) 0x4B, (byte) 0xE0, (byte) 0xF0, 
		(byte) 0x4B, (byte) 0xA0, (byte) 0x00, (byte) 0xDE, (byte) 0x4E, (byte) 0xE0, (byte) 0xD0, (byte) 0x31, 
		(byte) 0xBD, (byte) 0x51, (byte) 0xE0, (byte) 0x1D, (byte) 0x54, (byte) 0xE0, (byte) 0xD0, (byte) 0x1B, 
		(byte) 0xBD, (byte) 0x5D, (byte) 0xE0, (byte) 0x9D, (byte) 0x57, (byte) 0xE0, (byte) 0x9D, (byte) 0x4E, 
		(byte) 0xE0, (byte) 0xBD, (byte) 0x4B, (byte) 0xE0, (byte) 0x0A, (byte) 0xBD, (byte) 0x5A, (byte) 0xE0, 
		(byte) 0x90, (byte) 0x04, (byte) 0x49, (byte) 0xFF, (byte) 0x69, (byte) 0x00, (byte) 0x9D, (byte) 0x4B, 
		(byte) 0xE0, (byte) 0xD0, (byte) 0x10, (byte) 0xBD, (byte) 0x57, (byte) 0xE0, (byte) 0x9D, (byte) 0x4E, 
		(byte) 0xE0, (byte) 0x98, (byte) 0x38, (byte) 0xFD, (byte) 0x4B, (byte) 0xE0, (byte) 0x9D, (byte) 0x4B, 
		(byte) 0xE0, (byte) 0xC9, (byte) 0x00, (byte) 0x10, (byte) 0x01, (byte) 0x88, (byte) 0x18, (byte) 0x7D, 
		(byte) 0x51, (byte) 0xE0, (byte) 0x9D, (byte) 0x51, (byte) 0xE0, (byte) 0x98, (byte) 0x7D, (byte) 0x54, 
		(byte) 0xE0, (byte) 0x9D, (byte) 0x54, (byte) 0xE0, (byte) 0xBD, (byte) 0x36, (byte) 0xE0, (byte) 0x30, 
		(byte) 0x15, (byte) 0xBD, (byte) 0x93, (byte) 0xE0, (byte) 0xF0, (byte) 0x10, (byte) 0x18, (byte) 0x7D, 
		(byte) 0x14, (byte) 0xE0, (byte) 0x9D, (byte) 0x14, (byte) 0xE0, (byte) 0xBD, (byte) 0x96, (byte) 0xE0, 
		(byte) 0x7D, (byte) 0x17, (byte) 0xE0, (byte) 0x9D, (byte) 0x17, (byte) 0xE0, (byte) 0xBD, (byte) 0x63, 
		(byte) 0xE0, (byte) 0xF0, (byte) 0x4B, (byte) 0xA0, (byte) 0x00, (byte) 0xDE, (byte) 0x66, (byte) 0xE0, 
		(byte) 0xD0, (byte) 0x31, (byte) 0xBD, (byte) 0x69, (byte) 0xE0, (byte) 0x1D, (byte) 0x6C, (byte) 0xE0, 
		(byte) 0xD0, (byte) 0x1B, (byte) 0xBD, (byte) 0x72, (byte) 0xE0, (byte) 0x9D, (byte) 0x6F, (byte) 0xE0, 
		(byte) 0x9D, (byte) 0x66, (byte) 0xE0, (byte) 0xBD, (byte) 0x63, (byte) 0xE0, (byte) 0x0A, (byte) 0xBD, 
		(byte) 0x75, (byte) 0xE0, (byte) 0x90, (byte) 0x04, (byte) 0x49, (byte) 0xFF, (byte) 0x69, (byte) 0x00, 
		(byte) 0x9D, (byte) 0x63, (byte) 0xE0, (byte) 0xD0, (byte) 0x10, (byte) 0xBD, (byte) 0x6F, (byte) 0xE0, 
		(byte) 0x9D, (byte) 0x66, (byte) 0xE0, (byte) 0x98, (byte) 0x38, (byte) 0xFD, (byte) 0x63, (byte) 0xE0, 
		(byte) 0x9D, (byte) 0x63, (byte) 0xE0, (byte) 0xC9, (byte) 0x00, (byte) 0x10, (byte) 0x01, (byte) 0x88, 
		(byte) 0x18, (byte) 0x7D, (byte) 0x69, (byte) 0xE0, (byte) 0x9D, (byte) 0x69, (byte) 0xE0, (byte) 0x98, 
		(byte) 0x7D, (byte) 0x6C, (byte) 0xE0, (byte) 0x9D, (byte) 0x6C, (byte) 0xE0, (byte) 0xBD, (byte) 0x36, 
		(byte) 0xE0, (byte) 0x10, (byte) 0x03, (byte) 0x4C, (byte) 0x9F, (byte) 0xE6, (byte) 0xA0, (byte) 0x00, 
		(byte) 0xBD, (byte) 0xA2, (byte) 0xE0, (byte) 0xF0, (byte) 0x1C, (byte) 0x10, (byte) 0x01, (byte) 0xC8, 
		(byte) 0x18, (byte) 0x6D, (byte) 0x23, (byte) 0xE0, (byte) 0x48, (byte) 0x29, (byte) 0x07, (byte) 0x8D, 
		(byte) 0x23, (byte) 0xE0, (byte) 0x68, (byte) 0x6A, (byte) 0x4A, (byte) 0x4A, (byte) 0x18, (byte) 0x79, 
		(byte) 0xA6, (byte) 0xE1, (byte) 0x18, (byte) 0x6D, (byte) 0x24, (byte) 0xE0, (byte) 0x8D, (byte) 0x24, 
		(byte) 0xE0, (byte) 0x60, (byte) 0xBD, (byte) 0xA8, (byte) 0xE0, (byte) 0x85, (byte) 0xFD, (byte) 0xBD, 
		(byte) 0xAB, (byte) 0xE0, (byte) 0x85, (byte) 0xFE, (byte) 0xD0, (byte) 0x04, (byte) 0x60, (byte) 0x20, 
		(byte) 0x98, (byte) 0xE8, (byte) 0xAD, (byte) 0x00, (byte) 0xE0, (byte) 0x3D, (byte) 0x62, (byte) 0xE1, 
		(byte) 0xF0, (byte) 0xF4, (byte) 0xA0, (byte) 0x00, (byte) 0xB1, (byte) 0xFD, (byte) 0x85, (byte) 0xFF, 
		(byte) 0xC8, (byte) 0xB1, (byte) 0xFD, (byte) 0xA8, (byte) 0xA5, (byte) 0xFD, (byte) 0x18, (byte) 0x69, 
		(byte) 0x02, (byte) 0x85, (byte) 0xFD, (byte) 0x9D, (byte) 0xA8, (byte) 0xE0, (byte) 0xA5, (byte) 0xFE, 
		(byte) 0x69, (byte) 0x00, (byte) 0x85, (byte) 0xFE, (byte) 0x9D, (byte) 0xAB, (byte) 0xE0, (byte) 0xA5, 
		(byte) 0xFF, (byte) 0x29, (byte) 0x03, (byte) 0xD0, (byte) 0xD2, (byte) 0xBD, (byte) 0x8D, (byte) 0xE0, 
		(byte) 0x9D, (byte) 0x0E, (byte) 0xE0, (byte) 0xBD, (byte) 0x90, (byte) 0xE0, (byte) 0x9D, (byte) 0x11, 
		(byte) 0xE0, (byte) 0xA5, (byte) 0xFF, (byte) 0x9D, (byte) 0x05, (byte) 0xE0, (byte) 0x98, (byte) 0x9D, 
		(byte) 0x02, (byte) 0xE0, (byte) 0x29, (byte) 0x07, (byte) 0xA8, (byte) 0xB9, (byte) 0x67, (byte) 0xE1, 
		(byte) 0x8D, (byte) 0x6F, (byte) 0xE1, (byte) 0xBD, (byte) 0x02, (byte) 0xE0, (byte) 0x29, (byte) 0x38, 
		(byte) 0x4A, (byte) 0x4A, (byte) 0x4A, (byte) 0x7D, (byte) 0x81, (byte) 0xE0, (byte) 0x85, (byte) 0xFD, 
		(byte) 0xBD, (byte) 0x02, (byte) 0xE0, (byte) 0x29, (byte) 0xC0, (byte) 0x0A, (byte) 0x2A, (byte) 0x2A, 
		(byte) 0xA8, (byte) 0xB9, (byte) 0x6F, (byte) 0xE1, (byte) 0x85, (byte) 0xFE, (byte) 0xBD, (byte) 0x02, 
		(byte) 0xE0, (byte) 0x29, (byte) 0x07, (byte) 0xF0, (byte) 0x62, (byte) 0xA8, (byte) 0xB9, (byte) 0x72, 
		(byte) 0xE1, (byte) 0x65, (byte) 0xFE, (byte) 0x18, (byte) 0x7D, (byte) 0x84, (byte) 0xE0, (byte) 0x10, 
		(byte) 0x05, (byte) 0x18, (byte) 0x69, (byte) 0x0C, (byte) 0xE6, (byte) 0xFD, (byte) 0xC9, (byte) 0x0C, 
		(byte) 0x90, (byte) 0x04, (byte) 0xE9, (byte) 0x0C, (byte) 0xC6, (byte) 0xFD, (byte) 0x85, (byte) 0xFE, 
		(byte) 0xA8, (byte) 0xB9, (byte) 0x86, (byte) 0xE1, (byte) 0x85, (byte) 0xFF, (byte) 0xB9, (byte) 0x7A, 
		(byte) 0xE1, (byte) 0xA4, (byte) 0xFD, (byte) 0x88, (byte) 0x30, (byte) 0x06, (byte) 0x46, (byte) 0xFF, 
		(byte) 0x6A, (byte) 0x88, (byte) 0x10, (byte) 0xFA, (byte) 0x18, (byte) 0x7D, (byte) 0x87, (byte) 0xE0, 
		(byte) 0x9D, (byte) 0x8D, (byte) 0xE0, (byte) 0xA5, (byte) 0xFF, (byte) 0x7D, (byte) 0x8A, (byte) 0xE0, 
		(byte) 0x9D, (byte) 0x90, (byte) 0xE0, (byte) 0xBD, (byte) 0x05, (byte) 0xE0, (byte) 0xD0, (byte) 0x03, 
		(byte) 0x4C, (byte) 0xA0, (byte) 0xE6, (byte) 0xBD, (byte) 0x45, (byte) 0xE0, (byte) 0x1D, (byte) 0x48, 
		(byte) 0xE0, (byte) 0xF0, (byte) 0x16, (byte) 0xBD, (byte) 0x0E, (byte) 0xE0, (byte) 0xDD, (byte) 0x8D, 
		(byte) 0xE0, (byte) 0xBD, (byte) 0x11, (byte) 0xE0, (byte) 0xFD, (byte) 0x90, (byte) 0xE0, (byte) 0xA9, 
		(byte) 0xFE, (byte) 0x6A, (byte) 0x9D, (byte) 0x42, (byte) 0xE0, (byte) 0x90, (byte) 0x11, (byte) 0xF0, 
		(byte) 0x4A, (byte) 0x9D, (byte) 0x42, (byte) 0xE0, (byte) 0xBD, (byte) 0x8D, (byte) 0xE0, (byte) 0x9D, 
		(byte) 0x0E, (byte) 0xE0, (byte) 0xBD, (byte) 0x90, (byte) 0xE0, (byte) 0x9D, (byte) 0x11, (byte) 0xE0, 
		(byte) 0xBD, (byte) 0x36, (byte) 0xE0, (byte) 0x0A, (byte) 0xD0, (byte) 0x35, (byte) 0xBD, (byte) 0x93, 
		(byte) 0xE0, (byte) 0xF0, (byte) 0x0C, (byte) 0xBD, (byte) 0x99, (byte) 0xE0, (byte) 0x9D, (byte) 0x14, 
		(byte) 0xE0, (byte) 0xBD, (byte) 0x9C, (byte) 0xE0, (byte) 0x9D, (byte) 0x17, (byte) 0xE0, (byte) 0xBD, 
		(byte) 0x9F, (byte) 0xE0, (byte) 0xF0, (byte) 0x0F, (byte) 0xA4, (byte) 0xFD, (byte) 0x18, (byte) 0x79, 
		(byte) 0x92, (byte) 0xE1, (byte) 0xA4, (byte) 0xFE, (byte) 0x18, (byte) 0x79, (byte) 0x9A, (byte) 0xE1, 
		(byte) 0x18, (byte) 0x90, (byte) 0x08, (byte) 0xBD, (byte) 0xA2, (byte) 0xE0, (byte) 0xF0, (byte) 0x0B, 
		(byte) 0xBD, (byte) 0xA5, (byte) 0xE0, (byte) 0x8D, (byte) 0x24, (byte) 0xE0, (byte) 0xA9, (byte) 0x00, 
		(byte) 0x8D, (byte) 0x23, (byte) 0xE0, (byte) 0xBD, (byte) 0x3C, (byte) 0xE0, (byte) 0x9D, (byte) 0x3F, 
		(byte) 0xE0, (byte) 0xBD, (byte) 0x05, (byte) 0xE0, (byte) 0x29, (byte) 0x40, (byte) 0x9D, (byte) 0x36, 
		(byte) 0xE0, (byte) 0xBD, (byte) 0x05, (byte) 0xE0, (byte) 0x4A, (byte) 0x4A, (byte) 0x29, (byte) 0x07, 
		(byte) 0xD0, (byte) 0x30, (byte) 0xBD, (byte) 0x05, (byte) 0xE0, (byte) 0x30, (byte) 0x14, (byte) 0xAD, 
		(byte) 0x27, (byte) 0xE0, (byte) 0x29, (byte) 0x3C, (byte) 0xD0, (byte) 0x1E, (byte) 0xAD, (byte) 0x27, 
		(byte) 0xE0, (byte) 0x0A, (byte) 0x2A, (byte) 0x2A, (byte) 0xD0, (byte) 0x02, (byte) 0xA9, (byte) 0x04, 
		(byte) 0x4C, (byte) 0x70, (byte) 0xE8, (byte) 0xAD, (byte) 0x28, (byte) 0xE0, (byte) 0xF0, (byte) 0x0C, 
		(byte) 0x29, (byte) 0x3F, (byte) 0xD0, (byte) 0x08, (byte) 0xAD, (byte) 0x28, (byte) 0xE0, (byte) 0x0A, 
		(byte) 0x2A, (byte) 0x2A, (byte) 0xD0, (byte) 0x66, (byte) 0xA9, (byte) 0x10, (byte) 0x8D, (byte) 0x00, 
		(byte) 0xE0, (byte) 0x60, (byte) 0xC9, (byte) 0x01, (byte) 0xD0, (byte) 0x13, (byte) 0xBD, (byte) 0x05, 
		(byte) 0xE0, (byte) 0x29, (byte) 0x20, (byte) 0xD0, (byte) 0x06, (byte) 0xAD, (byte) 0x29, (byte) 0xE0, 
		(byte) 0x4C, (byte) 0x70, (byte) 0xE8, (byte) 0xBD, (byte) 0x2A, (byte) 0xE0, (byte) 0x4C, (byte) 0x70, 
		(byte) 0xE8, (byte) 0xA8, (byte) 0xBD, (byte) 0x05, (byte) 0xE0, (byte) 0x29, (byte) 0xA0, (byte) 0xC9, 
		(byte) 0x80, (byte) 0xF0, (byte) 0x30, (byte) 0x85, (byte) 0xFF, (byte) 0x18, (byte) 0xAD, (byte) 0x27, 
		(byte) 0xE0, (byte) 0xD0, (byte) 0x01, (byte) 0x38, (byte) 0x88, (byte) 0x88, (byte) 0xF0, (byte) 0x06, 
		(byte) 0x6A, (byte) 0xB0, (byte) 0x4E, (byte) 0x88, (byte) 0xD0, (byte) 0xFA, (byte) 0xA4, (byte) 0xFF, 
		(byte) 0x85, (byte) 0xFF, (byte) 0xF0, (byte) 0x26, (byte) 0x46, (byte) 0xFF, (byte) 0xB0, (byte) 0x41, 
		(byte) 0xF0, (byte) 0x42, (byte) 0x65, (byte) 0xFF, (byte) 0xB0, (byte) 0x3E, (byte) 0xC8, (byte) 0x10, 
		(byte) 0x19, (byte) 0x46, (byte) 0xFF, (byte) 0xB0, (byte) 0x34, (byte) 0x65, (byte) 0xFF, (byte) 0x90, 
		(byte) 0x11, (byte) 0xB0, (byte) 0x31, (byte) 0xAD, (byte) 0x28, (byte) 0xE0, (byte) 0xF0, (byte) 0x29, 
		(byte) 0x88, (byte) 0x88, (byte) 0xF0, (byte) 0x06, (byte) 0x4A, (byte) 0xB0, (byte) 0x22, (byte) 0x88, 
		(byte) 0xD0, (byte) 0xFA, (byte) 0x9D, (byte) 0x30, (byte) 0xE0, (byte) 0xBD, (byte) 0x1A, (byte) 0xE0, 
		(byte) 0x29, (byte) 0xF6, (byte) 0x9D, (byte) 0x1A, (byte) 0xE0, (byte) 0x38, (byte) 0xBD, (byte) 0x02, 
		(byte) 0xE0, (byte) 0x29, (byte) 0x07, (byte) 0xD0, (byte) 0x03, (byte) 0x7E, (byte) 0x36, (byte) 0xE0, 
		(byte) 0xBD, (byte) 0x1A, (byte) 0xE0, (byte) 0x69, (byte) 0x00, (byte) 0x9D, (byte) 0x1A, (byte) 0xE0, 
		(byte) 0x60, (byte) 0xA9, (byte) 0x10, (byte) 0x2C, (byte) 0xA9, (byte) 0x18, (byte) 0x8D, (byte) 0x00, 
		(byte) 0xE0, (byte) 0x60, (byte) 0x98, (byte) 0x48, (byte) 0xA5, (byte) 0xFF, (byte) 0x4A, (byte) 0x90, 
		(byte) 0x03, (byte) 0x4C, (byte) 0x42, (byte) 0xEA, (byte) 0x4A, (byte) 0x4A, (byte) 0xB0, (byte) 0x1E, 
		(byte) 0x4A, (byte) 0xB0, (byte) 0x0E, (byte) 0x9D, (byte) 0x9C, (byte) 0xE0, (byte) 0x9D, (byte) 0x17, 
		(byte) 0xE0, (byte) 0x68, (byte) 0x9D, (byte) 0x99, (byte) 0xE0, (byte) 0x9D, (byte) 0x14, (byte) 0xE0, 
		(byte) 0x60, (byte) 0x4A, (byte) 0x90, (byte) 0x02, (byte) 0x09, (byte) 0xF8, (byte) 0x9D, (byte) 0x8A, 
		(byte) 0xE0, (byte) 0x68, (byte) 0x9D, (byte) 0x87, (byte) 0xE0, (byte) 0x60, (byte) 0x4A, (byte) 0xB0, 
		(byte) 0x03, (byte) 0x4C, (byte) 0x4A, (byte) 0xE9, (byte) 0x4A, (byte) 0xB0, (byte) 0x61, (byte) 0x4A, 
		(byte) 0xB0, (byte) 0x0F, (byte) 0xD0, (byte) 0x08, (byte) 0x68, (byte) 0x9D, (byte) 0xA5, (byte) 0xE0, 
		(byte) 0x8D, (byte) 0x24, (byte) 0xE0, (byte) 0x60, (byte) 0x68, (byte) 0x9D, (byte) 0x3C, (byte) 0xE0, 
		(byte) 0x60, (byte) 0xD0, (byte) 0x48, (byte) 0x68, (byte) 0x9D, (byte) 0x7E, (byte) 0xE0, (byte) 0xC9, 
		(byte) 0x5B, (byte) 0xF0, (byte) 0x33, (byte) 0xA8, (byte) 0x4A, (byte) 0x4A, (byte) 0x4A, (byte) 0x38, 
		(byte) 0xE9, (byte) 0x0B, (byte) 0x18, (byte) 0x7D, (byte) 0x84, (byte) 0xE0, (byte) 0x30, (byte) 0x0C, 
		(byte) 0xC9, (byte) 0x0C, (byte) 0x90, (byte) 0x11, (byte) 0xE9, (byte) 0x0C, (byte) 0xDE, (byte) 0x81, 
		(byte) 0xE0, (byte) 0x4C, (byte) 0x0B, (byte) 0xE9, (byte) 0xC9, (byte) 0xF5, (byte) 0xB0, (byte) 0x05, 
		(byte) 0x69, (byte) 0x0C, (byte) 0xFE, (byte) 0x81, (byte) 0xE0, (byte) 0x9D, (byte) 0x84, (byte) 0xE0, 
		(byte) 0x98, (byte) 0x29, (byte) 0x07, (byte) 0x38, (byte) 0xE9, (byte) 0x03, (byte) 0x18, (byte) 0x7D, 
		(byte) 0x81, (byte) 0xE0, (byte) 0x9D, (byte) 0x81, (byte) 0xE0, (byte) 0x60, (byte) 0xBD, (byte) 0x78, 
		(byte) 0xE0, (byte) 0x9D, (byte) 0x81, (byte) 0xE0, (byte) 0xBD, (byte) 0x7B, (byte) 0xE0, (byte) 0x9D, 
		(byte) 0x84, (byte) 0xE0, (byte) 0x60, (byte) 0x68, (byte) 0x9D, (byte) 0xC6, (byte) 0xE0, (byte) 0x60, 
		(byte) 0x4A, (byte) 0xB0, (byte) 0x08, (byte) 0x9D, (byte) 0x0B, (byte) 0xE0, (byte) 0x68, (byte) 0x9D, 
		(byte) 0x08, (byte) 0xE0, (byte) 0x60, (byte) 0x4A, (byte) 0x6A, (byte) 0x6A, (byte) 0x6D, (byte) 0x5B, 
		(byte) 0xE1, (byte) 0x8D, (byte) 0x2D, (byte) 0xE0, (byte) 0x68, (byte) 0x6D, (byte) 0x5C, (byte) 0xE1, 
		(byte) 0x8D, (byte) 0x2E, (byte) 0xE0, (byte) 0x60, (byte) 0x4A, (byte) 0x90, (byte) 0x03, (byte) 0x4C, 
		(byte) 0xD3, (byte) 0xE9, (byte) 0x4A, (byte) 0xB0, (byte) 0x40, (byte) 0x4A, (byte) 0xB0, (byte) 0x17, 
		(byte) 0x4A, (byte) 0xB0, (byte) 0x0F, (byte) 0x68, (byte) 0x8D, (byte) 0x27, (byte) 0xE0, (byte) 0x4A, 
		(byte) 0x4A, (byte) 0x4A, (byte) 0xA8, (byte) 0xB9, (byte) 0xAF, (byte) 0xE1, (byte) 0x8D, (byte) 0x28, 
		(byte) 0xE0, (byte) 0x60, (byte) 0x68, (byte) 0x9D, (byte) 0x5D, (byte) 0xE0, (byte) 0x60, (byte) 0x4A, 
		(byte) 0xB0, (byte) 0x05, (byte) 0x68, (byte) 0x8D, (byte) 0x01, (byte) 0xE0, (byte) 0x60, (byte) 0x68, 
		(byte) 0xF0, (byte) 0x11, (byte) 0x9D, (byte) 0x75, (byte) 0xE0, (byte) 0xBC, (byte) 0x63, (byte) 0xE0, 
		(byte) 0xD0, (byte) 0x08, (byte) 0x9D, (byte) 0x63, (byte) 0xE0, (byte) 0xA9, (byte) 0x01, (byte) 0x9D, 
		(byte) 0x66, (byte) 0xE0, (byte) 0x60, (byte) 0x9D, (byte) 0x63, (byte) 0xE0, (byte) 0x9D, (byte) 0x69, 
		(byte) 0xE0, (byte) 0x9D, (byte) 0x6C, (byte) 0xE0, (byte) 0x60, (byte) 0x4A, (byte) 0xB0, (byte) 0x30, 
		(byte) 0x4A, (byte) 0xB0, (byte) 0x05, (byte) 0x68, (byte) 0x9D, (byte) 0x39, (byte) 0xE0, (byte) 0x60, 
		(byte) 0x68, (byte) 0xA0, (byte) 0x00, (byte) 0x4A, (byte) 0x90, (byte) 0x02, (byte) 0xC8, (byte) 0x18, 
		(byte) 0x48, (byte) 0x29, (byte) 0x07, (byte) 0x79, (byte) 0xAC, (byte) 0xE1, (byte) 0x9D, (byte) 0x78, 
		(byte) 0xE0, (byte) 0x9D, (byte) 0x81, (byte) 0xE0, (byte) 0x68, (byte) 0x4A, (byte) 0x4A, (byte) 0x4A, 
		(byte) 0x18, (byte) 0x79, (byte) 0xAD, (byte) 0xE1, (byte) 0x9D, (byte) 0x7B, (byte) 0xE0, (byte) 0x9D, 
		(byte) 0x84, (byte) 0xE0, (byte) 0xA9, (byte) 0x5B, (byte) 0x9D, (byte) 0x7E, (byte) 0xE0, (byte) 0x60, 
		(byte) 0x4A, (byte) 0xB0, (byte) 0x05, (byte) 0x68, (byte) 0x9D, (byte) 0xA2, (byte) 0xE0, (byte) 0x60, 
		(byte) 0x68, (byte) 0x8D, (byte) 0xCC, (byte) 0xE0, (byte) 0x60, (byte) 0x4A, (byte) 0xB0, (byte) 0x27, 
		(byte) 0x4A, (byte) 0xB0, (byte) 0x0D, (byte) 0x4A, (byte) 0xB0, (byte) 0x05, (byte) 0x68, (byte) 0x8D, 
		(byte) 0x29, (byte) 0xE0, (byte) 0x60, (byte) 0x68, (byte) 0x9D, (byte) 0x9F, (byte) 0xE0, (byte) 0x60, 
		(byte) 0x4A, (byte) 0xB0, (byte) 0x0F, (byte) 0x68, (byte) 0x9D, (byte) 0x93, (byte) 0xE0, (byte) 0xA0, 
		(byte) 0x00, (byte) 0x0A, (byte) 0x90, (byte) 0x01, (byte) 0x88, (byte) 0x98, (byte) 0x9D, (byte) 0x96, 
		(byte) 0xE0, (byte) 0x60, (byte) 0x68, (byte) 0x9D, (byte) 0x72, (byte) 0xE0, (byte) 0x60, (byte) 0x4A, 
		(byte) 0xB0, (byte) 0x1C, (byte) 0x4A, (byte) 0xB0, (byte) 0x15, (byte) 0x68, (byte) 0x9D, (byte) 0xB7, 
		(byte) 0xE0, (byte) 0xA5, (byte) 0xFD, (byte) 0x9D, (byte) 0xB1, (byte) 0xE0, (byte) 0xA5, (byte) 0xFE, 
		(byte) 0x9D, (byte) 0xB4, (byte) 0xE0, (byte) 0xBD, (byte) 0x33, (byte) 0xE0, (byte) 0x9D, (byte) 0xAE, 
		(byte) 0xE0, (byte) 0x60, (byte) 0x68, (byte) 0x6C, (byte) 0x5F, (byte) 0xE1, (byte) 0x4A, (byte) 0xB0, 
		(byte) 0x1E, (byte) 0x68, (byte) 0xD0, (byte) 0x0A, (byte) 0x9D, (byte) 0x4B, (byte) 0xE0, (byte) 0x9D, 
		(byte) 0x51, (byte) 0xE0, (byte) 0x9D, (byte) 0x54, (byte) 0xE0, (byte) 0x60, (byte) 0x9D, (byte) 0x5A, 
		(byte) 0xE0, (byte) 0xBC, (byte) 0x4B, (byte) 0xE0, (byte) 0xD0, (byte) 0x08, (byte) 0x9D, (byte) 0x4B, 
		(byte) 0xE0, (byte) 0xA9, (byte) 0x01, (byte) 0x9D, (byte) 0x4E, (byte) 0xE0, (byte) 0x60, (byte) 0x68, 
		(byte) 0x9D, (byte) 0x2A, (byte) 0xE0, (byte) 0x60, (byte) 0x4A, (byte) 0x90, (byte) 0x08, (byte) 0x9D, 
		(byte) 0x48, (byte) 0xE0, (byte) 0x68, (byte) 0x9D, (byte) 0x45, (byte) 0xE0, (byte) 0x60, (byte) 0x68, 
		(byte) 0x4A, (byte) 0xB0, (byte) 0x61, (byte) 0x4A, (byte) 0xB0, (byte) 0x25, (byte) 0x4A, (byte) 0xB0, 
		(byte) 0x05, (byte) 0x4A, (byte) 0xA0, (byte) 0xF0, (byte) 0xD0, (byte) 0x06, (byte) 0x0A, (byte) 0x0A, 
		(byte) 0x0A, (byte) 0x0A, (byte) 0xA0, (byte) 0x0F, (byte) 0x85, (byte) 0xFF, (byte) 0x98, (byte) 0xB0, 
		(byte) 0x09, (byte) 0x3D, (byte) 0x1D, (byte) 0xE0, (byte) 0x05, (byte) 0xFF, (byte) 0x9D, (byte) 0x1D, 
		(byte) 0xE0, (byte) 0x60, (byte) 0x3D, (byte) 0x20, (byte) 0xE0, (byte) 0x05, (byte) 0xFF, (byte) 0x9D, 
		(byte) 0x20, (byte) 0xE0, (byte) 0x60, (byte) 0x4A, (byte) 0xB0, (byte) 0x38, (byte) 0x4A, (byte) 0xB0, 
		(byte) 0x64, (byte) 0x85, (byte) 0xFF, (byte) 0xBD, (byte) 0xBA, (byte) 0xE0, (byte) 0xDD, (byte) 0xA9, 
		(byte) 0xE1, (byte) 0xF0, (byte) 0x54, (byte) 0xFE, (byte) 0xBA, (byte) 0xE0, (byte) 0xA8, (byte) 0xA5, 
		(byte) 0xFD, (byte) 0x99, (byte) 0xE1, (byte) 0xE0, (byte) 0xA5, (byte) 0xFE, (byte) 0x99, (byte) 0xF0, 
		(byte) 0xE0, (byte) 0xBD, (byte) 0x33, (byte) 0xE0, (byte) 0x99, (byte) 0x2F, (byte) 0xE1, (byte) 0xA4, 
		(byte) 0xFF, (byte) 0xB9, (byte) 0x17, (byte) 0xE1, (byte) 0xF0, (byte) 0x36, (byte) 0x85, (byte) 0xFE, 
		(byte) 0xB9, (byte) 0xFF, (byte) 0xE0, (byte) 0x85, (byte) 0xFD, (byte) 0xB9, (byte) 0x3E, (byte) 0xE1, 
		(byte) 0x9D, (byte) 0x33, (byte) 0xE0, (byte) 0x60, (byte) 0xB0, (byte) 0x4B, (byte) 0x4A, (byte) 0xB0, 
		(byte) 0x3C, (byte) 0xA8, (byte) 0xA5, (byte) 0xFD, (byte) 0x99, (byte) 0xFF, (byte) 0xE0, (byte) 0xA5, 
		(byte) 0xFE, (byte) 0x99, (byte) 0x17, (byte) 0xE1, (byte) 0xBD, (byte) 0x33, (byte) 0xE0, (byte) 0x99, 
		(byte) 0x3E, (byte) 0xE1, (byte) 0xBD, (byte) 0xBA, (byte) 0xE0, (byte) 0xDD, (byte) 0xA9, (byte) 0xE1, 
		(byte) 0xF0, (byte) 0x0D, (byte) 0xFE, (byte) 0xBA, (byte) 0xE0, (byte) 0xA8, (byte) 0xA9, (byte) 0x00, 
		(byte) 0x99, (byte) 0xF0, (byte) 0xE0, (byte) 0x60, (byte) 0xA9, (byte) 0x30, (byte) 0x2C, (byte) 0xA9, 
		(byte) 0x28, (byte) 0x8D, (byte) 0x00, (byte) 0xE0, (byte) 0x60, (byte) 0x0A, (byte) 0x0A, (byte) 0x0A, 
		(byte) 0x0A, (byte) 0x4D, (byte) 0x25, (byte) 0xE0, (byte) 0x29, (byte) 0xF0, (byte) 0x4D, (byte) 0x25, 
		(byte) 0xE0, (byte) 0x8D, (byte) 0x25, (byte) 0xE0, (byte) 0x60, (byte) 0x4D, (byte) 0x26, (byte) 0xE0, 
		(byte) 0x29, (byte) 0x0F, (byte) 0x4D, (byte) 0x26, (byte) 0xE0, (byte) 0x8D, (byte) 0x26, (byte) 0xE0, 
		(byte) 0x60, (byte) 0x4A, (byte) 0xB0, (byte) 0x0B, (byte) 0x4A, (byte) 0xB0, (byte) 0x04, (byte) 0x8D, 
		(byte) 0xCA, (byte) 0xE0, (byte) 0x60, (byte) 0x8D, (byte) 0xCB, (byte) 0xE0, (byte) 0x60, (byte) 0x4A, 
		(byte) 0x90, (byte) 0x03, (byte) 0x4C, (byte) 0xA5, (byte) 0xEB, (byte) 0x4A, (byte) 0xA8, (byte) 0xF0, 
		(byte) 0x21, (byte) 0x88, (byte) 0xF0, (byte) 0x34, (byte) 0x88, (byte) 0xF0, (byte) 0x42, (byte) 0x88, 
		(byte) 0xF0, (byte) 0x4A, (byte) 0x88, (byte) 0xF0, (byte) 0x52, (byte) 0x88, (byte) 0xF0, (byte) 0x5C, 
		(byte) 0x88, (byte) 0xF0, (byte) 0x66, (byte) 0x88, (byte) 0xF0, (byte) 0x73, (byte) 0x29, (byte) 0x07, 
		(byte) 0x09, (byte) 0x10, (byte) 0xB0, (byte) 0x03, (byte) 0x4C, (byte) 0xB7, (byte) 0xEA, (byte) 0x4C, 
		(byte) 0x7F, (byte) 0xEA, (byte) 0xAC, (byte) 0x26, (byte) 0xE0, (byte) 0xB0, (byte) 0x07, (byte) 0xC8, 
		(byte) 0x98, (byte) 0x29, (byte) 0x0F, (byte) 0xD0, (byte) 0x07, (byte) 0x60, (byte) 0x98, (byte) 0x29, 
		(byte) 0x0F, (byte) 0xF0, (byte) 0x04, (byte) 0x88, (byte) 0x8C, (byte) 0x26, (byte) 0xE0, (byte) 0x60, 
		(byte) 0xBD, (byte) 0x62, (byte) 0xE1, (byte) 0x49, (byte) 0xFF, (byte) 0x2D, (byte) 0x25, (byte) 0xE0, 
		(byte) 0x90, (byte) 0x03, (byte) 0x1D, (byte) 0x62, (byte) 0xE1, (byte) 0x8D, (byte) 0x25, (byte) 0xE0, 
		(byte) 0x60, (byte) 0xBD, (byte) 0x1A, (byte) 0xE0, (byte) 0x29, (byte) 0xFB, (byte) 0x90, (byte) 0x55, 
		(byte) 0x09, (byte) 0x04, (byte) 0xB0, (byte) 0x51, (byte) 0xBD, (byte) 0x1A, (byte) 0xE0, (byte) 0x29, 
		(byte) 0xFD, (byte) 0x90, (byte) 0x4A, (byte) 0x09, (byte) 0x02, (byte) 0xB0, (byte) 0x46, (byte) 0xAD, 
		(byte) 0x25, (byte) 0xE0, (byte) 0x29, (byte) 0xF7, (byte) 0x90, (byte) 0x02, (byte) 0x09, (byte) 0x08, 
		(byte) 0x8D, (byte) 0x25, (byte) 0xE0, (byte) 0x60, (byte) 0xAD, (byte) 0x26, (byte) 0xE0, (byte) 0x29, 
		(byte) 0x7F, (byte) 0x90, (byte) 0x02, (byte) 0x09, (byte) 0x80, (byte) 0x8D, (byte) 0x26, (byte) 0xE0, 
		(byte) 0x60, (byte) 0x98, (byte) 0x8D, (byte) 0xBD, (byte) 0xE0, (byte) 0x8D, (byte) 0xDF, (byte) 0xE0, 
		(byte) 0xC8, (byte) 0x8C, (byte) 0xE0, (byte) 0xE0, (byte) 0x2A, (byte) 0x8D, (byte) 0xC9, (byte) 0xE0, 
		(byte) 0x60, (byte) 0x98, (byte) 0x2A, (byte) 0x9D, (byte) 0x60, (byte) 0xE0, (byte) 0x60, (byte) 0x4A, 
		(byte) 0xB0, (byte) 0x27, (byte) 0x4A, (byte) 0xB0, (byte) 0x14, (byte) 0xD0, (byte) 0x02, (byte) 0xA9, 
		(byte) 0x08, (byte) 0x0A, (byte) 0x0A, (byte) 0x0A, (byte) 0x0A, (byte) 0x5D, (byte) 0x1A, (byte) 0xE0, 
		(byte) 0x29, (byte) 0xF0, (byte) 0x5D, (byte) 0x1A, (byte) 0xE0, (byte) 0x9D, (byte) 0x1A, (byte) 0xE0, 
		(byte) 0x60, (byte) 0x0A, (byte) 0x0A, (byte) 0x0A, (byte) 0x0A, (byte) 0x4D, (byte) 0x26, (byte) 0xE0, 
		(byte) 0x29, (byte) 0x70, (byte) 0x4D, (byte) 0x26, (byte) 0xE0, (byte) 0x8D, (byte) 0x26, (byte) 0xE0, 
		(byte) 0x60, (byte) 0x4A, (byte) 0x90, (byte) 0x04, (byte) 0x9D, (byte) 0xC0, (byte) 0xE0, (byte) 0x60, 
		(byte) 0xA8, (byte) 0xF0, (byte) 0x20, (byte) 0x88, (byte) 0xF0, (byte) 0x40, (byte) 0x88, (byte) 0xF0, 
		(byte) 0x63, (byte) 0x29, (byte) 0x03, (byte) 0x9D, (byte) 0xC3, (byte) 0xE0, (byte) 0xA9, (byte) 0x00, 
		(byte) 0x9D, (byte) 0xCD, (byte) 0xE0, (byte) 0x9D, (byte) 0xD0, (byte) 0xE0, (byte) 0x9D, (byte) 0xD3, 
		(byte) 0xE0, (byte) 0x9D, (byte) 0xD6, (byte) 0xE0, (byte) 0x8D, (byte) 0xD9, (byte) 0xE0, (byte) 0x8D, 
		(byte) 0xDC, (byte) 0xE0, (byte) 0x60, (byte) 0xBD, (byte) 0xB7, (byte) 0xE0, (byte) 0xF0, (byte) 0x05, 
		(byte) 0xDE, (byte) 0xB7, (byte) 0xE0, (byte) 0xF0, (byte) 0x12, (byte) 0xBD, (byte) 0x33, (byte) 0xE0, 
		(byte) 0xDD, (byte) 0xAE, (byte) 0xE0, (byte) 0xD0, (byte) 0x0B, (byte) 0xBD, (byte) 0xB1, (byte) 0xE0, 
		(byte) 0x85, (byte) 0xFD, (byte) 0xBD, (byte) 0xB4, (byte) 0xE0, (byte) 0x85, (byte) 0xFE, (byte) 0x60, 
		(byte) 0xA9, (byte) 0x38, (byte) 0x8D, (byte) 0x00, (byte) 0xE0, (byte) 0x60, (byte) 0xBD, (byte) 0xBA, 
		(byte) 0xE0, (byte) 0xDD, (byte) 0xA8, (byte) 0xE1, (byte) 0xF0, (byte) 0x18, (byte) 0xDE, (byte) 0xBA, 
		(byte) 0xE0, (byte) 0xA8, (byte) 0x88, (byte) 0xB9, (byte) 0xF0, (byte) 0xE0, (byte) 0xF0, (byte) 0x0D, 
		(byte) 0x85, (byte) 0xFE, (byte) 0xB9, (byte) 0xE1, (byte) 0xE0, (byte) 0x85, (byte) 0xFD, (byte) 0xB9, 
		(byte) 0x2F, (byte) 0xE1, (byte) 0x9D, (byte) 0x33, (byte) 0xE0, (byte) 0x60, (byte) 0xA9, (byte) 0x20, 
		(byte) 0x8D, (byte) 0x00, (byte) 0xE0, (byte) 0x60, (byte) 0xAD, (byte) 0x00, (byte) 0xE0, (byte) 0x5D, 
		(byte) 0x62, (byte) 0xE1, (byte) 0x8D, (byte) 0x00, (byte) 0xE0, (byte) 0xA9, (byte) 0x01, (byte) 0x9D, 
		(byte) 0x30, (byte) 0xE0, (byte) 0x60, (byte) 0xAD, (byte) 0x00, (byte) 0xE0, (byte) 0x29, (byte) 0x07, 
		(byte) 0x8D, (byte) 0x81, (byte) 0xEC, (byte) 0xD0, (byte) 0x03, (byte) 0x20, (byte) 0xE9, (byte) 0xE2, 
		(byte) 0x60, (byte) 0x00, (byte) 0xA2, (byte) 0x51, (byte) 0xA0, (byte) 0xEC, (byte) 0x8E, (byte) 0x5D, 
		(byte) 0xE1, (byte) 0x8C, (byte) 0x5E, (byte) 0xE1, (byte) 0x20, (byte) 0xCF, (byte) 0xE1, (byte) 0xA2, 
		(byte) 0x00, (byte) 0xA0, (byte) 0x09, (byte) 0x20, (byte) 0x00, (byte) 0xE2, (byte) 0xA9, (byte) 0x07, 
		(byte) 0x8D, (byte) 0x00, (byte) 0xE0, (byte) 0x8D, (byte) 0x81, (byte) 0xEC, (byte) 0x60, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0xA9, (byte) 0x00, (byte) 0x29, (byte) 0xFF, (byte) 0xF0, (byte) 0xF6, 
		(byte) 0x4C, (byte) 0x29, (byte) 0xE3, (byte) 0xA9, (byte) 0x07, (byte) 0x8D, (byte) 0x00, (byte) 0xE0, 
		(byte) 0x60, 
	};

	private static final byte _sidtune_sidplayer2[] = {

		(byte) 0x00, (byte) 0xF0, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
		(byte) 0x00, (byte) 0x61, (byte) 0xF1, (byte) 0x60, (byte) 0x01, (byte) 0x02, (byte) 0x04, (byte) 0x00, 
		(byte) 0x07, (byte) 0x0E, (byte) 0x02, (byte) 0x02, (byte) 0xFE, (byte) 0x02, (byte) 0x02, (byte) 0xFE, 
		(byte) 0xFE, (byte) 0x00, (byte) 0x01, (byte) 0x00, (byte) 0xFF, (byte) 0x00, (byte) 0x02, (byte) 0x04, 
		(byte) 0x05, (byte) 0x07, (byte) 0x09, (byte) 0x0B, (byte) 0x1E, (byte) 0x18, (byte) 0x8B, (byte) 0x7E, 
		(byte) 0xFA, (byte) 0x06, (byte) 0xAC, (byte) 0xF3, (byte) 0xE6, (byte) 0x8F, (byte) 0xF8, (byte) 0x2E, 
		(byte) 0x86, (byte) 0x8E, (byte) 0x96, (byte) 0x9F, (byte) 0xA8, (byte) 0xB3, (byte) 0xBD, (byte) 0xC8, 
		(byte) 0xD4, (byte) 0xE1, (byte) 0xEE, (byte) 0xFD, (byte) 0x8C, (byte) 0x78, (byte) 0x64, (byte) 0x50, 
		(byte) 0x3C, (byte) 0x28, (byte) 0x14, (byte) 0x00, (byte) 0x00, (byte) 0x02, (byte) 0x03, (byte) 0x05, 
		(byte) 0x07, (byte) 0x08, (byte) 0x0A, (byte) 0x0C, (byte) 0x0D, (byte) 0x0F, (byte) 0x11, (byte) 0x12, 
		(byte) 0x00, (byte) 0xE0, (byte) 0x00, (byte) 0x05, (byte) 0x0A, (byte) 0x0F, (byte) 0xF9, (byte) 0x00, 
		(byte) 0xF5, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x10, (byte) 0x00, (byte) 0x00, (byte) 0x20, 
		(byte) 0x00, (byte) 0x00, (byte) 0x30, (byte) 0x00, (byte) 0x00, (byte) 0x40, (byte) 0x00, (byte) 0x00, 
		(byte) 0x50, (byte) 0x00, (byte) 0x00, (byte) 0x60, (byte) 0x00, (byte) 0x00, (byte) 0x70, (byte) 0x00, 
		(byte) 0x00, (byte) 0x80, (byte) 0x00, (byte) 0x00, (byte) 0x90, (byte) 0x00, (byte) 0x00, (byte) 0xA0, 
		(byte) 0x00, (byte) 0xA9, (byte) 0x00, (byte) 0x8D, (byte) 0x00, (byte) 0xF0, (byte) 0xA2, (byte) 0x95, 
		(byte) 0xA0, (byte) 0x42, (byte) 0xAD, (byte) 0xA6, (byte) 0x02, (byte) 0xF0, (byte) 0x04, (byte) 0xA2, 
		(byte) 0x25, (byte) 0xA0, (byte) 0x40, (byte) 0x8E, (byte) 0x5B, (byte) 0xF1, (byte) 0x8C, (byte) 0x5C, 
		(byte) 0xF1, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, 
		(byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, 
		(byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, 
		(byte) 0xEA, (byte) 0x60, (byte) 0xA9, (byte) 0x00, (byte) 0x8D, (byte) 0x00, (byte) 0xF0, (byte) 0x86, 
		(byte) 0x61, (byte) 0x84, (byte) 0x62, (byte) 0xA0, (byte) 0xBC, (byte) 0x99, (byte) 0x00, (byte) 0xF0, 
		(byte) 0x88, (byte) 0xD0, (byte) 0xFA, (byte) 0xA0, (byte) 0x72, (byte) 0x99, (byte) 0xBC, (byte) 0xF0, 
		(byte) 0x88, (byte) 0xD0, (byte) 0xFA, (byte) 0x8D, (byte) 0x15, (byte) 0xD5, (byte) 0x8D, (byte) 0x16, 
		(byte) 0xD5, (byte) 0xA9, (byte) 0x08, (byte) 0x8D, (byte) 0x25, (byte) 0xF0, (byte) 0x8D, (byte) 0x17, 
		(byte) 0xD5, (byte) 0x8D, (byte) 0x26, (byte) 0xF0, (byte) 0x8D, (byte) 0x18, (byte) 0xD5, (byte) 0xA9, 
		(byte) 0x90, (byte) 0x8D, (byte) 0x27, (byte) 0xF0, (byte) 0xA9, (byte) 0x60, (byte) 0x8D, (byte) 0x28, 
		(byte) 0xF0, (byte) 0xA9, (byte) 0x0C, (byte) 0x8D, (byte) 0x29, (byte) 0xF0, (byte) 0xAD, (byte) 0x5B, 
		(byte) 0xF1, (byte) 0x8D, (byte) 0x2D, (byte) 0xF0, (byte) 0xAD, (byte) 0x5C, (byte) 0xF1, (byte) 0x8D, 
		(byte) 0x2E, (byte) 0xF0, (byte) 0xA9, (byte) 0xFF, (byte) 0x8D, (byte) 0xCC, (byte) 0xF0, (byte) 0xA9, 
		(byte) 0xD5, (byte) 0x85, (byte) 0x64, (byte) 0xA2, (byte) 0x02, (byte) 0xA9, (byte) 0xFF, (byte) 0x9D, 
		(byte) 0x0B, (byte) 0xF0, (byte) 0xA9, (byte) 0x01, (byte) 0x9D, (byte) 0x30, (byte) 0xF0, (byte) 0x9D, 
		(byte) 0x2A, (byte) 0xF0, (byte) 0x8A, (byte) 0x9D, (byte) 0x33, (byte) 0xF0, (byte) 0x9D, (byte) 0xAE, 
		(byte) 0xF0, (byte) 0xA9, (byte) 0x04, (byte) 0x9D, (byte) 0x39, (byte) 0xF0, (byte) 0xBD, (byte) 0xA8, 
		(byte) 0xF1, (byte) 0x9D, (byte) 0xBA, (byte) 0xF0, (byte) 0xA9, (byte) 0x5B, (byte) 0x9D, (byte) 0x7E, 
		(byte) 0xF0, (byte) 0xBD, (byte) 0x65, (byte) 0xF1, (byte) 0x85, (byte) 0x63, (byte) 0xA9, (byte) 0x00, 
		(byte) 0xA8, (byte) 0x91, (byte) 0x63, (byte) 0xC8, (byte) 0x91, (byte) 0x63, (byte) 0xC8, (byte) 0x91, 
		(byte) 0x63, (byte) 0xA9, (byte) 0x08, (byte) 0x9D, (byte) 0x17, (byte) 0xF0, (byte) 0x9D, (byte) 0x9C, 
		(byte) 0xF0, (byte) 0xC8, (byte) 0x91, (byte) 0x63, (byte) 0xC8, (byte) 0x91, (byte) 0x63, (byte) 0xA9, 
		(byte) 0x40, (byte) 0x9D, (byte) 0x1A, (byte) 0xF0, (byte) 0x91, (byte) 0x63, (byte) 0xA9, (byte) 0x20, 
		(byte) 0x9D, (byte) 0x1D, (byte) 0xF0, (byte) 0xC8, (byte) 0x91, (byte) 0x63, (byte) 0xA9, (byte) 0xF5, 
		(byte) 0x9D, (byte) 0x20, (byte) 0xF0, (byte) 0xC8, (byte) 0x91, (byte) 0x63, (byte) 0xCA, (byte) 0x10, 
		(byte) 0xA4, (byte) 0x8A, (byte) 0xA2, (byte) 0x17, (byte) 0x9D, (byte) 0x3E, (byte) 0xF1, (byte) 0xCA, 
		(byte) 0x10, (byte) 0xFA, (byte) 0xA5, (byte) 0x61, (byte) 0x18, (byte) 0x69, (byte) 0x06, (byte) 0x85, 
		(byte) 0x63, (byte) 0xA9, (byte) 0x00, (byte) 0xAA, (byte) 0xA8, (byte) 0x65, (byte) 0x62, (byte) 0x85, 
		(byte) 0x64, (byte) 0x9D, (byte) 0xAB, (byte) 0xF0, (byte) 0x9D, (byte) 0xB4, (byte) 0xF0, (byte) 0xA5, 
		(byte) 0x63, (byte) 0x9D, (byte) 0xA8, (byte) 0xF0, (byte) 0x9D, (byte) 0xB1, (byte) 0xF0, (byte) 0x18, 
		(byte) 0x71, (byte) 0x61, (byte) 0x85, (byte) 0x63, (byte) 0xA5, (byte) 0x64, (byte) 0xC8, (byte) 0x71, 
		(byte) 0x61, (byte) 0xC8, (byte) 0xE8, (byte) 0xE0, (byte) 0x03, (byte) 0xD0, (byte) 0xE0, (byte) 0xA6, 
		(byte) 0x63, (byte) 0xA8, (byte) 0x60, (byte) 0xA9, (byte) 0x00, (byte) 0x8D, (byte) 0x04, (byte) 0xD5, 
		(byte) 0x8D, (byte) 0x0B, (byte) 0xD5, (byte) 0x8D, (byte) 0x12, (byte) 0xD5, (byte) 0x8D, (byte) 0x01, 
		(byte) 0xD5, (byte) 0x8D, (byte) 0x08, (byte) 0xD5, (byte) 0x8D, (byte) 0x0F, (byte) 0xD5, (byte) 0xA9, 
		(byte) 0x08, (byte) 0x8D, (byte) 0x17, (byte) 0xD5, (byte) 0xAD, (byte) 0x5B, (byte) 0xF1, (byte) 0x8D, 
		(byte) 0x04, (byte) 0xDC, (byte) 0xAD, (byte) 0x5C, (byte) 0xF1, (byte) 0x8D, (byte) 0x05, (byte) 0xDC, 
		(byte) 0x60, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, 
		(byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0x60, 
		(byte) 0xA9, (byte) 0x08, (byte) 0x8D, (byte) 0x00, (byte) 0xF0, (byte) 0x6C, (byte) 0x5D, (byte) 0xF1, 
		(byte) 0xEA, (byte) 0xEA, (byte) 0xEA, (byte) 0xAD, (byte) 0x00, (byte) 0xF0, (byte) 0x30, (byte) 0xF0, 
		(byte) 0x09, (byte) 0x80, (byte) 0xA8, (byte) 0x29, (byte) 0x07, (byte) 0xF0, (byte) 0xEE, (byte) 0xD8, 
		(byte) 0x8C, (byte) 0x00, (byte) 0xF0, (byte) 0xEA, (byte) 0xA5, (byte) 0xFB, (byte) 0x8D, (byte) 0x56, 
		(byte) 0xF1, (byte) 0xA5, (byte) 0xFC, (byte) 0x8D, (byte) 0x57, (byte) 0xF1, (byte) 0xA5, (byte) 0xFD, 
		(byte) 0x8D, (byte) 0x58, (byte) 0xF1, (byte) 0xA5, (byte) 0xFE, (byte) 0x8D, (byte) 0x59, (byte) 0xF1, 
		(byte) 0xA5, (byte) 0xFF, (byte) 0x8D, (byte) 0x5A, (byte) 0xF1, (byte) 0xAD, (byte) 0x23, (byte) 0xF0, 
		(byte) 0x18, (byte) 0x6D, (byte) 0xD9, (byte) 0xF0, (byte) 0x48, (byte) 0x29, (byte) 0x07, (byte) 0xA8, 
		(byte) 0xAD, (byte) 0xDC, (byte) 0xF0, (byte) 0x69, (byte) 0x00, (byte) 0x85, (byte) 0xFF, (byte) 0x68, 
		(byte) 0x46, (byte) 0xFF, (byte) 0x6A, (byte) 0x46, (byte) 0xFF, (byte) 0x6A, (byte) 0x46, (byte) 0xFF, 
		(byte) 0x6A, (byte) 0x18, (byte) 0x6D, (byte) 0x24, (byte) 0xF0, (byte) 0x8C, (byte) 0x15, (byte) 0xD5, 
		(byte) 0x8D, (byte) 0x16, (byte) 0xD5, (byte) 0xAD, (byte) 0x25, (byte) 0xF0, (byte) 0x8D, (byte) 0x17, 
		(byte) 0xD5, (byte) 0xAD, (byte) 0x26, (byte) 0xF0, (byte) 0x8D, (byte) 0x18, (byte) 0xD5, (byte) 0xA9, 
		(byte) 0xD5, (byte) 0x85, (byte) 0xFC, (byte) 0xA2, (byte) 0x00, (byte) 0xAD, (byte) 0x00, (byte) 0xF0, 
		(byte) 0x3D, (byte) 0x62, (byte) 0xF1, (byte) 0xF0, (byte) 0x51, (byte) 0xBD, (byte) 0x65, (byte) 0xF1, 
		(byte) 0x85, (byte) 0xFB, (byte) 0xBD, (byte) 0x0E, (byte) 0xF0, (byte) 0x18, (byte) 0x7D, (byte) 0x51, 
		(byte) 0xF0, (byte) 0xA8, (byte) 0xBD, (byte) 0x11, (byte) 0xF0, (byte) 0x7D, (byte) 0x54, (byte) 0xF0, 
		(byte) 0x48, (byte) 0x98, (byte) 0x18, (byte) 0x7D, (byte) 0xCD, (byte) 0xF0, (byte) 0xA0, (byte) 0x00, 
		(byte) 0x91, (byte) 0xFB, (byte) 0x68, (byte) 0x7D, (byte) 0xD0, (byte) 0xF0, (byte) 0xC8, (byte) 0x91, 
		(byte) 0xFB, (byte) 0xBD, (byte) 0x14, (byte) 0xF0, (byte) 0x18, (byte) 0x7D, (byte) 0x69, (byte) 0xF0, 
		(byte) 0x85, (byte) 0xFF, (byte) 0xBD, (byte) 0x17, (byte) 0xF0, (byte) 0x7D, (byte) 0x6C, (byte) 0xF0, 
		(byte) 0x48, (byte) 0xA5, (byte) 0xFF, (byte) 0x18, (byte) 0x7D, (byte) 0xD3, (byte) 0xF0, (byte) 0xC8, 
		(byte) 0x91, (byte) 0xFB, (byte) 0x68, (byte) 0x7D, (byte) 0xD6, (byte) 0xF0, (byte) 0xC8, (byte) 0x91, 
		(byte) 0xFB, (byte) 0xBD, (byte) 0x1D, (byte) 0xF0, (byte) 0xC8, (byte) 0xC8, (byte) 0x91, (byte) 0xFB, 
		(byte) 0xBD, (byte) 0x20, (byte) 0xF0, (byte) 0xC8, (byte) 0x91, (byte) 0xFB, (byte) 0xE8, (byte) 0xE0, 
		(byte) 0x03, (byte) 0xD0, (byte) 0xA2, (byte) 0xAC, (byte) 0x1A, (byte) 0xF0, (byte) 0xAE, (byte) 0x1B, 
		(byte) 0xF0, (byte) 0xAD, (byte) 0x1C, (byte) 0xF0, (byte) 0x8C, (byte) 0x04, (byte) 0xD5, (byte) 0x8E, 
		(byte) 0x0B, (byte) 0xD5, (byte) 0x8D, (byte) 0x12, (byte) 0xD5, (byte) 0xAE, (byte) 0x2D, (byte) 0xF0, 
		(byte) 0xAC, (byte) 0x2E, (byte) 0xF0, (byte) 0x8E, (byte) 0x04, (byte) 0xDC, (byte) 0x8C, (byte) 0x05, 
		(byte) 0xDC, (byte) 0xAD, (byte) 0x1B, (byte) 0xD5, (byte) 0x8D, (byte) 0xBE, (byte) 0xF0, (byte) 0xAD, 
		(byte) 0x1C, (byte) 0xD5, (byte) 0x8D, (byte) 0xBF, (byte) 0xF0, (byte) 0xA2, (byte) 0x00, (byte) 0xAD, 
		(byte) 0x00, (byte) 0xF0, (byte) 0x3D, (byte) 0x62, (byte) 0xF1, (byte) 0xF0, (byte) 0x10, (byte) 0x8E, 
		(byte) 0x2F, (byte) 0xF0, (byte) 0x20, (byte) 0x36, (byte) 0xF5, (byte) 0xAD, (byte) 0x00, (byte) 0xF0, 
		(byte) 0x29, (byte) 0x78, (byte) 0xF0, (byte) 0x03, (byte) 0x4C, (byte) 0x0C, (byte) 0xF5, (byte) 0xE8, 
		(byte) 0xE0, (byte) 0x03, (byte) 0xD0, (byte) 0xE3, (byte) 0xAD, (byte) 0xC9, (byte) 0xF0, (byte) 0xD0, 
		(byte) 0x52, (byte) 0xAD, (byte) 0xCA, (byte) 0xF0, (byte) 0x0D, (byte) 0xCB, (byte) 0xF0, (byte) 0xF0, 
		(byte) 0x78, (byte) 0xAD, (byte) 0xDF, (byte) 0xF0, (byte) 0xD0, (byte) 0x28, (byte) 0xAD, (byte) 0xCA, 
		(byte) 0xF0, (byte) 0xF0, (byte) 0x28, (byte) 0x18, (byte) 0x6D, (byte) 0xBD, (byte) 0xF0, (byte) 0xB0, 
		(byte) 0x07, (byte) 0xCD, (byte) 0xCC, (byte) 0xF0, (byte) 0x90, (byte) 0x60, (byte) 0xF0, (byte) 0x5E, 
		(byte) 0xA9, (byte) 0x00, (byte) 0x8D, (byte) 0xDF, (byte) 0xF0, (byte) 0xAD, (byte) 0xCB, (byte) 0xF0, 
		(byte) 0xF0, (byte) 0x54, (byte) 0xEE, (byte) 0xDF, (byte) 0xF0, (byte) 0xAD, (byte) 0xBD, (byte) 0xF0, 
		(byte) 0xED, (byte) 0xCB, (byte) 0xF0, (byte) 0x4C, (byte) 0xB4, (byte) 0xF4, (byte) 0xAD, (byte) 0xCB, 
		(byte) 0xF0, (byte) 0xF0, (byte) 0xD3, (byte) 0xAD, (byte) 0xBD, (byte) 0xF0, (byte) 0x38, (byte) 0xED, 
		(byte) 0xCB, (byte) 0xF0, (byte) 0xB0, (byte) 0x3A, (byte) 0xA9, (byte) 0x00, (byte) 0x8D, (byte) 0xDF, 
		(byte) 0xF0, (byte) 0xAD, (byte) 0xCA, (byte) 0xF0, (byte) 0xD0, (byte) 0x30, (byte) 0xEE, (byte) 0xDF, 
		(byte) 0xF0, (byte) 0xD0, (byte) 0x28, (byte) 0xCE, (byte) 0xE0, (byte) 0xF0, (byte) 0xD0, (byte) 0x29, 
		(byte) 0xAD, (byte) 0xDF, (byte) 0xF0, (byte) 0xD0, (byte) 0x11, (byte) 0xEE, (byte) 0xDF, (byte) 0xF0, 
		(byte) 0xAD, (byte) 0xCB, (byte) 0xF0, (byte) 0xD0, (byte) 0x02, (byte) 0xA9, (byte) 0x20, (byte) 0x8D, 
		(byte) 0xE0, (byte) 0xF0, (byte) 0xA9, (byte) 0x00, (byte) 0xF0, (byte) 0x10, (byte) 0xCE, (byte) 0xDF, 
		(byte) 0xF0, (byte) 0xAD, (byte) 0xCA, (byte) 0xF0, (byte) 0xD0, (byte) 0x02, (byte) 0xA9, (byte) 0x20, 
		(byte) 0x8D, (byte) 0xE0, (byte) 0xF0, (byte) 0xAD, (byte) 0xCC, (byte) 0xF0, (byte) 0x8D, (byte) 0xBD, 
		(byte) 0xF0, (byte) 0xA2, (byte) 0x00, (byte) 0xBD, (byte) 0xC3, (byte) 0xF0, (byte) 0xF0, (byte) 0x44, 
		(byte) 0xA9, (byte) 0x00, (byte) 0x85, (byte) 0xFF, (byte) 0xBC, (byte) 0xC0, (byte) 0xF0, (byte) 0xB9, 
		(byte) 0xBD, (byte) 0xF0, (byte) 0xBC, (byte) 0xC6, (byte) 0xF0, (byte) 0xF0, (byte) 0x0E, (byte) 0x30, 
		(byte) 0x08, (byte) 0x0A, (byte) 0x26, (byte) 0xFF, (byte) 0x88, (byte) 0xD0, (byte) 0xFA, (byte) 0xF0, 
		(byte) 0x04, (byte) 0x4A, (byte) 0xC8, (byte) 0xD0, (byte) 0xFC, (byte) 0xBC, (byte) 0xC3, (byte) 0xF0, 
		(byte) 0x88, (byte) 0xD0, (byte) 0x0B, (byte) 0x9D, (byte) 0xCD, (byte) 0xF0, (byte) 0xA5, (byte) 0xFF, 
		(byte) 0x9D, (byte) 0xD0, (byte) 0xF0, (byte) 0x4C, (byte) 0x02, (byte) 0xF5, (byte) 0x88, (byte) 0xD0, 
		(byte) 0x0B, (byte) 0x9D, (byte) 0xD3, (byte) 0xF0, (byte) 0xA5, (byte) 0xFF, (byte) 0x9D, (byte) 0xD6, 
		(byte) 0xF0, (byte) 0x4C, (byte) 0x02, (byte) 0xF5, (byte) 0x8D, (byte) 0xD9, (byte) 0xF0, (byte) 0xA5, 
		(byte) 0xFF, (byte) 0x8D, (byte) 0xDC, (byte) 0xF0, (byte) 0xE8, (byte) 0xE0, (byte) 0x03, (byte) 0xD0, 
		(byte) 0xB2, (byte) 0xAD, (byte) 0x00, (byte) 0xF0, (byte) 0x29, (byte) 0x7F, (byte) 0x8D, (byte) 0x00, 
		(byte) 0xF0, (byte) 0xAD, (byte) 0x56, (byte) 0xF1, (byte) 0x85, (byte) 0xFB, (byte) 0xAD, (byte) 0x57, 
		(byte) 0xF1, (byte) 0x85, (byte) 0xFC, (byte) 0xAD, (byte) 0x58, (byte) 0xF1, (byte) 0x85, (byte) 0xFD, 
		(byte) 0xAD, (byte) 0x59, (byte) 0xF1, (byte) 0x85, (byte) 0xFE, (byte) 0xAD, (byte) 0x5A, (byte) 0xF1, 
		(byte) 0x85, (byte) 0xFF, (byte) 0x6C, (byte) 0x5D, (byte) 0xF1, (byte) 0xBD, (byte) 0x60, (byte) 0xF0, 
		(byte) 0xD0, (byte) 0x03, (byte) 0x4C, (byte) 0x9F, (byte) 0xF6, (byte) 0x4C, (byte) 0xBA, (byte) 0xF5, 
		(byte) 0xDE, (byte) 0x30, (byte) 0xF0, (byte) 0xD0, (byte) 0x03, (byte) 0x4C, (byte) 0xA0, (byte) 0xF6, 
		(byte) 0xBD, (byte) 0x36, (byte) 0xF0, (byte) 0x30, (byte) 0xE8, (byte) 0xD0, (byte) 0x1A, (byte) 0xBD, 
		(byte) 0x3F, (byte) 0xF0, (byte) 0xF0, (byte) 0x05, (byte) 0xDE, (byte) 0x3F, (byte) 0xF0, (byte) 0xD0, 
		(byte) 0x10, (byte) 0xBD, (byte) 0x39, (byte) 0xF0, (byte) 0xDD, (byte) 0x30, (byte) 0xF0, (byte) 0x90, 
		(byte) 0x08, (byte) 0xBD, (byte) 0x1A, (byte) 0xF0, (byte) 0x29, (byte) 0xFE, (byte) 0x9D, (byte) 0x1A, 
		(byte) 0xF0, (byte) 0xBD, (byte) 0x42, (byte) 0xF0, (byte) 0xF0, (byte) 0x56, (byte) 0x0A, (byte) 0xBD, 
		(byte) 0x0E, (byte) 0xF0, (byte) 0xB0, (byte) 0x1D, (byte) 0x7D, (byte) 0x45, (byte) 0xF0, (byte) 0x9D, 
		(byte) 0x0E, (byte) 0xF0, (byte) 0xA8, (byte) 0xBD, (byte) 0x11, (byte) 0xF0, (byte) 0x7D, (byte) 0x48, 
		(byte) 0xF0, (byte) 0x9D, (byte) 0x11, (byte) 0xF0, (byte) 0x48, (byte) 0x98, (byte) 0xDD, (byte) 0x8D, 
		(byte) 0xF0, (byte) 0x68, (byte) 0xFD, (byte) 0x90, (byte) 0xF0, (byte) 0xB0, (byte) 0x1F, (byte) 0x90, 
		(byte) 0x2E, (byte) 0xFD, (byte) 0x45, (byte) 0xF0, (byte) 0x9D, (byte) 0x0E, (byte) 0xF0, (byte) 0xBD, 
		(byte) 0x11, (byte) 0xF0, (byte) 0xFD, (byte) 0x48, (byte) 0xF0, (byte) 0x9D, (byte) 0x11, (byte) 0xF0, 
		(byte) 0xBD, (byte) 0x8D, (byte) 0xF0, (byte) 0xDD, (byte) 0x0E, (byte) 0xF0, (byte) 0xBD, (byte) 0x90, 
		(byte) 0xF0, (byte) 0xFD, (byte) 0x11, (byte) 0xF0, (byte) 0x90, (byte) 0x11, (byte) 0xBD, (byte) 0x8D, 
		(byte) 0xF0, (byte) 0x9D, (byte) 0x0E, (byte) 0xF0, (byte) 0xBD, (byte) 0x90, (byte) 0xF0, (byte) 0x9D, 
		(byte) 0x11, (byte) 0xF0, (byte) 0xA9, (byte) 0x00, (byte) 0x9D, (byte) 0x42, (byte) 0xF0, (byte) 0xBD, 
		(byte) 0x60, (byte) 0xF0, (byte) 0xF0, (byte) 0x55, (byte) 0xBD, (byte) 0x4B, (byte) 0xF0, (byte) 0xF0, 
		(byte) 0x4B, (byte) 0xA0, (byte) 0x00, (byte) 0xDE, (byte) 0x4E, (byte) 0xF0, (byte) 0xD0, (byte) 0x31, 
		(byte) 0xBD, (byte) 0x51, (byte) 0xF0, (byte) 0x1D, (byte) 0x54, (byte) 0xF0, (byte) 0xD0, (byte) 0x1B, 
		(byte) 0xBD, (byte) 0x5D, (byte) 0xF0, (byte) 0x9D, (byte) 0x57, (byte) 0xF0, (byte) 0x9D, (byte) 0x4E, 
		(byte) 0xF0, (byte) 0xBD, (byte) 0x4B, (byte) 0xF0, (byte) 0x0A, (byte) 0xBD, (byte) 0x5A, (byte) 0xF0, 
		(byte) 0x90, (byte) 0x04, (byte) 0x49, (byte) 0xFF, (byte) 0x69, (byte) 0x00, (byte) 0x9D, (byte) 0x4B, 
		(byte) 0xF0, (byte) 0xD0, (byte) 0x10, (byte) 0xBD, (byte) 0x57, (byte) 0xF0, (byte) 0x9D, (byte) 0x4E, 
		(byte) 0xF0, (byte) 0x98, (byte) 0x38, (byte) 0xFD, (byte) 0x4B, (byte) 0xF0, (byte) 0x9D, (byte) 0x4B, 
		(byte) 0xF0, (byte) 0xC9, (byte) 0x00, (byte) 0x10, (byte) 0x01, (byte) 0x88, (byte) 0x18, (byte) 0x7D, 
		(byte) 0x51, (byte) 0xF0, (byte) 0x9D, (byte) 0x51, (byte) 0xF0, (byte) 0x98, (byte) 0x7D, (byte) 0x54, 
		(byte) 0xF0, (byte) 0x9D, (byte) 0x54, (byte) 0xF0, (byte) 0xBD, (byte) 0x36, (byte) 0xF0, (byte) 0x30, 
		(byte) 0x15, (byte) 0xBD, (byte) 0x93, (byte) 0xF0, (byte) 0xF0, (byte) 0x10, (byte) 0x18, (byte) 0x7D, 
		(byte) 0x14, (byte) 0xF0, (byte) 0x9D, (byte) 0x14, (byte) 0xF0, (byte) 0xBD, (byte) 0x96, (byte) 0xF0, 
		(byte) 0x7D, (byte) 0x17, (byte) 0xF0, (byte) 0x9D, (byte) 0x17, (byte) 0xF0, (byte) 0xBD, (byte) 0x63, 
		(byte) 0xF0, (byte) 0xF0, (byte) 0x4B, (byte) 0xA0, (byte) 0x00, (byte) 0xDE, (byte) 0x66, (byte) 0xF0, 
		(byte) 0xD0, (byte) 0x31, (byte) 0xBD, (byte) 0x69, (byte) 0xF0, (byte) 0x1D, (byte) 0x6C, (byte) 0xF0, 
		(byte) 0xD0, (byte) 0x1B, (byte) 0xBD, (byte) 0x72, (byte) 0xF0, (byte) 0x9D, (byte) 0x6F, (byte) 0xF0, 
		(byte) 0x9D, (byte) 0x66, (byte) 0xF0, (byte) 0xBD, (byte) 0x63, (byte) 0xF0, (byte) 0x0A, (byte) 0xBD, 
		(byte) 0x75, (byte) 0xF0, (byte) 0x90, (byte) 0x04, (byte) 0x49, (byte) 0xFF, (byte) 0x69, (byte) 0x00, 
		(byte) 0x9D, (byte) 0x63, (byte) 0xF0, (byte) 0xD0, (byte) 0x10, (byte) 0xBD, (byte) 0x6F, (byte) 0xF0, 
		(byte) 0x9D, (byte) 0x66, (byte) 0xF0, (byte) 0x98, (byte) 0x38, (byte) 0xFD, (byte) 0x63, (byte) 0xF0, 
		(byte) 0x9D, (byte) 0x63, (byte) 0xF0, (byte) 0xC9, (byte) 0x00, (byte) 0x10, (byte) 0x01, (byte) 0x88, 
		(byte) 0x18, (byte) 0x7D, (byte) 0x69, (byte) 0xF0, (byte) 0x9D, (byte) 0x69, (byte) 0xF0, (byte) 0x98, 
		(byte) 0x7D, (byte) 0x6C, (byte) 0xF0, (byte) 0x9D, (byte) 0x6C, (byte) 0xF0, (byte) 0xBD, (byte) 0x36, 
		(byte) 0xF0, (byte) 0x10, (byte) 0x03, (byte) 0x4C, (byte) 0x9F, (byte) 0xF6, (byte) 0xA0, (byte) 0x00, 
		(byte) 0xBD, (byte) 0xA2, (byte) 0xF0, (byte) 0xF0, (byte) 0x1C, (byte) 0x10, (byte) 0x01, (byte) 0xC8, 
		(byte) 0x18, (byte) 0x6D, (byte) 0x23, (byte) 0xF0, (byte) 0x48, (byte) 0x29, (byte) 0x07, (byte) 0x8D, 
		(byte) 0x23, (byte) 0xF0, (byte) 0x68, (byte) 0x6A, (byte) 0x4A, (byte) 0x4A, (byte) 0x18, (byte) 0x79, 
		(byte) 0xA6, (byte) 0xF1, (byte) 0x18, (byte) 0x6D, (byte) 0x24, (byte) 0xF0, (byte) 0x8D, (byte) 0x24, 
		(byte) 0xF0, (byte) 0x60, (byte) 0xBD, (byte) 0xA8, (byte) 0xF0, (byte) 0x85, (byte) 0xFD, (byte) 0xBD, 
		(byte) 0xAB, (byte) 0xF0, (byte) 0x85, (byte) 0xFE, (byte) 0xD0, (byte) 0x04, (byte) 0x60, (byte) 0x20, 
		(byte) 0x98, (byte) 0xF8, (byte) 0xAD, (byte) 0x00, (byte) 0xF0, (byte) 0x3D, (byte) 0x62, (byte) 0xF1, 
		(byte) 0xF0, (byte) 0xF4, (byte) 0xA0, (byte) 0x00, (byte) 0xB1, (byte) 0xFD, (byte) 0x85, (byte) 0xFF, 
		(byte) 0xC8, (byte) 0xB1, (byte) 0xFD, (byte) 0xA8, (byte) 0xA5, (byte) 0xFD, (byte) 0x18, (byte) 0x69, 
		(byte) 0x02, (byte) 0x85, (byte) 0xFD, (byte) 0x9D, (byte) 0xA8, (byte) 0xF0, (byte) 0xA5, (byte) 0xFE, 
		(byte) 0x69, (byte) 0x00, (byte) 0x85, (byte) 0xFE, (byte) 0x9D, (byte) 0xAB, (byte) 0xF0, (byte) 0xA5, 
		(byte) 0xFF, (byte) 0x29, (byte) 0x03, (byte) 0xD0, (byte) 0xD2, (byte) 0xBD, (byte) 0x8D, (byte) 0xF0, 
		(byte) 0x9D, (byte) 0x0E, (byte) 0xF0, (byte) 0xBD, (byte) 0x90, (byte) 0xF0, (byte) 0x9D, (byte) 0x11, 
		(byte) 0xF0, (byte) 0xA5, (byte) 0xFF, (byte) 0x9D, (byte) 0x05, (byte) 0xF0, (byte) 0x98, (byte) 0x9D, 
		(byte) 0x02, (byte) 0xF0, (byte) 0x29, (byte) 0x07, (byte) 0xA8, (byte) 0xB9, (byte) 0x67, (byte) 0xF1, 
		(byte) 0x8D, (byte) 0x6F, (byte) 0xF1, (byte) 0xBD, (byte) 0x02, (byte) 0xF0, (byte) 0x29, (byte) 0x38, 
		(byte) 0x4A, (byte) 0x4A, (byte) 0x4A, (byte) 0x7D, (byte) 0x81, (byte) 0xF0, (byte) 0x85, (byte) 0xFD, 
		(byte) 0xBD, (byte) 0x02, (byte) 0xF0, (byte) 0x29, (byte) 0xC0, (byte) 0x0A, (byte) 0x2A, (byte) 0x2A, 
		(byte) 0xA8, (byte) 0xB9, (byte) 0x6F, (byte) 0xF1, (byte) 0x85, (byte) 0xFE, (byte) 0xBD, (byte) 0x02, 
		(byte) 0xF0, (byte) 0x29, (byte) 0x07, (byte) 0xF0, (byte) 0x62, (byte) 0xA8, (byte) 0xB9, (byte) 0x72, 
		(byte) 0xF1, (byte) 0x65, (byte) 0xFE, (byte) 0x18, (byte) 0x7D, (byte) 0x84, (byte) 0xF0, (byte) 0x10, 
		(byte) 0x05, (byte) 0x18, (byte) 0x69, (byte) 0x0C, (byte) 0xE6, (byte) 0xFD, (byte) 0xC9, (byte) 0x0C, 
		(byte) 0x90, (byte) 0x04, (byte) 0xE9, (byte) 0x0C, (byte) 0xC6, (byte) 0xFD, (byte) 0x85, (byte) 0xFE, 
		(byte) 0xA8, (byte) 0xB9, (byte) 0x86, (byte) 0xF1, (byte) 0x85, (byte) 0xFF, (byte) 0xB9, (byte) 0x7A, 
		(byte) 0xF1, (byte) 0xA4, (byte) 0xFD, (byte) 0x88, (byte) 0x30, (byte) 0x06, (byte) 0x46, (byte) 0xFF, 
		(byte) 0x6A, (byte) 0x88, (byte) 0x10, (byte) 0xFA, (byte) 0x18, (byte) 0x7D, (byte) 0x87, (byte) 0xF0, 
		(byte) 0x9D, (byte) 0x8D, (byte) 0xF0, (byte) 0xA5, (byte) 0xFF, (byte) 0x7D, (byte) 0x8A, (byte) 0xF0, 
		(byte) 0x9D, (byte) 0x90, (byte) 0xF0, (byte) 0xBD, (byte) 0x05, (byte) 0xF0, (byte) 0xD0, (byte) 0x03, 
		(byte) 0x4C, (byte) 0xA0, (byte) 0xF6, (byte) 0xBD, (byte) 0x45, (byte) 0xF0, (byte) 0x1D, (byte) 0x48, 
		(byte) 0xF0, (byte) 0xF0, (byte) 0x16, (byte) 0xBD, (byte) 0x0E, (byte) 0xF0, (byte) 0xDD, (byte) 0x8D, 
		(byte) 0xF0, (byte) 0xBD, (byte) 0x11, (byte) 0xF0, (byte) 0xFD, (byte) 0x90, (byte) 0xF0, (byte) 0xA9, 
		(byte) 0xFE, (byte) 0x6A, (byte) 0x9D, (byte) 0x42, (byte) 0xF0, (byte) 0x90, (byte) 0x11, (byte) 0xF0, 
		(byte) 0x4A, (byte) 0x9D, (byte) 0x42, (byte) 0xF0, (byte) 0xBD, (byte) 0x8D, (byte) 0xF0, (byte) 0x9D, 
		(byte) 0x0E, (byte) 0xF0, (byte) 0xBD, (byte) 0x90, (byte) 0xF0, (byte) 0x9D, (byte) 0x11, (byte) 0xF0, 
		(byte) 0xBD, (byte) 0x36, (byte) 0xF0, (byte) 0x0A, (byte) 0xD0, (byte) 0x35, (byte) 0xBD, (byte) 0x93, 
		(byte) 0xF0, (byte) 0xF0, (byte) 0x0C, (byte) 0xBD, (byte) 0x99, (byte) 0xF0, (byte) 0x9D, (byte) 0x14, 
		(byte) 0xF0, (byte) 0xBD, (byte) 0x9C, (byte) 0xF0, (byte) 0x9D, (byte) 0x17, (byte) 0xF0, (byte) 0xBD, 
		(byte) 0x9F, (byte) 0xF0, (byte) 0xF0, (byte) 0x0F, (byte) 0xA4, (byte) 0xFD, (byte) 0x18, (byte) 0x79, 
		(byte) 0x92, (byte) 0xF1, (byte) 0xA4, (byte) 0xFE, (byte) 0x18, (byte) 0x79, (byte) 0x9A, (byte) 0xF1, 
		(byte) 0x18, (byte) 0x90, (byte) 0x08, (byte) 0xBD, (byte) 0xA2, (byte) 0xF0, (byte) 0xF0, (byte) 0x0B, 
		(byte) 0xBD, (byte) 0xA5, (byte) 0xF0, (byte) 0x8D, (byte) 0x24, (byte) 0xF0, (byte) 0xA9, (byte) 0x00, 
		(byte) 0x8D, (byte) 0x23, (byte) 0xF0, (byte) 0xBD, (byte) 0x3C, (byte) 0xF0, (byte) 0x9D, (byte) 0x3F, 
		(byte) 0xF0, (byte) 0xBD, (byte) 0x05, (byte) 0xF0, (byte) 0x29, (byte) 0x40, (byte) 0x9D, (byte) 0x36, 
		(byte) 0xF0, (byte) 0xBD, (byte) 0x05, (byte) 0xF0, (byte) 0x4A, (byte) 0x4A, (byte) 0x29, (byte) 0x07, 
		(byte) 0xD0, (byte) 0x30, (byte) 0xBD, (byte) 0x05, (byte) 0xF0, (byte) 0x30, (byte) 0x14, (byte) 0xAD, 
		(byte) 0x27, (byte) 0xF0, (byte) 0x29, (byte) 0x3C, (byte) 0xD0, (byte) 0x1E, (byte) 0xAD, (byte) 0x27, 
		(byte) 0xF0, (byte) 0x0A, (byte) 0x2A, (byte) 0x2A, (byte) 0xD0, (byte) 0x02, (byte) 0xA9, (byte) 0x04, 
		(byte) 0x4C, (byte) 0x70, (byte) 0xF8, (byte) 0xAD, (byte) 0x28, (byte) 0xF0, (byte) 0xF0, (byte) 0x0C, 
		(byte) 0x29, (byte) 0x3F, (byte) 0xD0, (byte) 0x08, (byte) 0xAD, (byte) 0x28, (byte) 0xF0, (byte) 0x0A, 
		(byte) 0x2A, (byte) 0x2A, (byte) 0xD0, (byte) 0x66, (byte) 0xA9, (byte) 0x10, (byte) 0x8D, (byte) 0x00, 
		(byte) 0xF0, (byte) 0x60, (byte) 0xC9, (byte) 0x01, (byte) 0xD0, (byte) 0x13, (byte) 0xBD, (byte) 0x05, 
		(byte) 0xF0, (byte) 0x29, (byte) 0x20, (byte) 0xD0, (byte) 0x06, (byte) 0xAD, (byte) 0x29, (byte) 0xF0, 
		(byte) 0x4C, (byte) 0x70, (byte) 0xF8, (byte) 0xBD, (byte) 0x2A, (byte) 0xF0, (byte) 0x4C, (byte) 0x70, 
		(byte) 0xF8, (byte) 0xA8, (byte) 0xBD, (byte) 0x05, (byte) 0xF0, (byte) 0x29, (byte) 0xA0, (byte) 0xC9, 
		(byte) 0x80, (byte) 0xF0, (byte) 0x30, (byte) 0x85, (byte) 0xFF, (byte) 0x18, (byte) 0xAD, (byte) 0x27, 
		(byte) 0xF0, (byte) 0xD0, (byte) 0x01, (byte) 0x38, (byte) 0x88, (byte) 0x88, (byte) 0xF0, (byte) 0x06, 
		(byte) 0x6A, (byte) 0xB0, (byte) 0x4E, (byte) 0x88, (byte) 0xD0, (byte) 0xFA, (byte) 0xA4, (byte) 0xFF, 
		(byte) 0x85, (byte) 0xFF, (byte) 0xF0, (byte) 0x26, (byte) 0x46, (byte) 0xFF, (byte) 0xB0, (byte) 0x41, 
		(byte) 0xF0, (byte) 0x42, (byte) 0x65, (byte) 0xFF, (byte) 0xB0, (byte) 0x3E, (byte) 0xC8, (byte) 0x10, 
		(byte) 0x19, (byte) 0x46, (byte) 0xFF, (byte) 0xB0, (byte) 0x34, (byte) 0x65, (byte) 0xFF, (byte) 0x90, 
		(byte) 0x11, (byte) 0xB0, (byte) 0x31, (byte) 0xAD, (byte) 0x28, (byte) 0xF0, (byte) 0xF0, (byte) 0x29, 
		(byte) 0x88, (byte) 0x88, (byte) 0xF0, (byte) 0x06, (byte) 0x4A, (byte) 0xB0, (byte) 0x22, (byte) 0x88, 
		(byte) 0xD0, (byte) 0xFA, (byte) 0x9D, (byte) 0x30, (byte) 0xF0, (byte) 0xBD, (byte) 0x1A, (byte) 0xF0, 
		(byte) 0x29, (byte) 0xF6, (byte) 0x9D, (byte) 0x1A, (byte) 0xF0, (byte) 0x38, (byte) 0xBD, (byte) 0x02, 
		(byte) 0xF0, (byte) 0x29, (byte) 0x07, (byte) 0xD0, (byte) 0x03, (byte) 0x7E, (byte) 0x36, (byte) 0xF0, 
		(byte) 0xBD, (byte) 0x1A, (byte) 0xF0, (byte) 0x69, (byte) 0x00, (byte) 0x9D, (byte) 0x1A, (byte) 0xF0, 
		(byte) 0x60, (byte) 0xA9, (byte) 0x10, (byte) 0x2C, (byte) 0xA9, (byte) 0x18, (byte) 0x8D, (byte) 0x00, 
		(byte) 0xF0, (byte) 0x60, (byte) 0x98, (byte) 0x48, (byte) 0xA5, (byte) 0xFF, (byte) 0x4A, (byte) 0x90, 
		(byte) 0x03, (byte) 0x4C, (byte) 0x42, (byte) 0xFA, (byte) 0x4A, (byte) 0x4A, (byte) 0xB0, (byte) 0x1E, 
		(byte) 0x4A, (byte) 0xB0, (byte) 0x0E, (byte) 0x9D, (byte) 0x9C, (byte) 0xF0, (byte) 0x9D, (byte) 0x17, 
		(byte) 0xF0, (byte) 0x68, (byte) 0x9D, (byte) 0x99, (byte) 0xF0, (byte) 0x9D, (byte) 0x14, (byte) 0xF0, 
		(byte) 0x60, (byte) 0x4A, (byte) 0x90, (byte) 0x02, (byte) 0x09, (byte) 0xF8, (byte) 0x9D, (byte) 0x8A, 
		(byte) 0xF0, (byte) 0x68, (byte) 0x9D, (byte) 0x87, (byte) 0xF0, (byte) 0x60, (byte) 0x4A, (byte) 0xB0, 
		(byte) 0x03, (byte) 0x4C, (byte) 0x4A, (byte) 0xF9, (byte) 0x4A, (byte) 0xB0, (byte) 0x61, (byte) 0x4A, 
		(byte) 0xB0, (byte) 0x0F, (byte) 0xD0, (byte) 0x08, (byte) 0x68, (byte) 0x9D, (byte) 0xA5, (byte) 0xF0, 
		(byte) 0x8D, (byte) 0x24, (byte) 0xF0, (byte) 0x60, (byte) 0x68, (byte) 0x9D, (byte) 0x3C, (byte) 0xF0, 
		(byte) 0x60, (byte) 0xD0, (byte) 0x48, (byte) 0x68, (byte) 0x9D, (byte) 0x7E, (byte) 0xF0, (byte) 0xC9, 
		(byte) 0x5B, (byte) 0xF0, (byte) 0x33, (byte) 0xA8, (byte) 0x4A, (byte) 0x4A, (byte) 0x4A, (byte) 0x38, 
		(byte) 0xE9, (byte) 0x0B, (byte) 0x18, (byte) 0x7D, (byte) 0x84, (byte) 0xF0, (byte) 0x30, (byte) 0x0C, 
		(byte) 0xC9, (byte) 0x0C, (byte) 0x90, (byte) 0x11, (byte) 0xE9, (byte) 0x0C, (byte) 0xDE, (byte) 0x81, 
		(byte) 0xF0, (byte) 0x4C, (byte) 0x0B, (byte) 0xF9, (byte) 0xC9, (byte) 0xF5, (byte) 0xB0, (byte) 0x05, 
		(byte) 0x69, (byte) 0x0C, (byte) 0xFE, (byte) 0x81, (byte) 0xF0, (byte) 0x9D, (byte) 0x84, (byte) 0xF0, 
		(byte) 0x98, (byte) 0x29, (byte) 0x07, (byte) 0x38, (byte) 0xE9, (byte) 0x03, (byte) 0x18, (byte) 0x7D, 
		(byte) 0x81, (byte) 0xF0, (byte) 0x9D, (byte) 0x81, (byte) 0xF0, (byte) 0x60, (byte) 0xBD, (byte) 0x78, 
		(byte) 0xF0, (byte) 0x9D, (byte) 0x81, (byte) 0xF0, (byte) 0xBD, (byte) 0x7B, (byte) 0xF0, (byte) 0x9D, 
		(byte) 0x84, (byte) 0xF0, (byte) 0x60, (byte) 0x68, (byte) 0x9D, (byte) 0xC6, (byte) 0xF0, (byte) 0x60, 
		(byte) 0x4A, (byte) 0xB0, (byte) 0x08, (byte) 0x9D, (byte) 0x0B, (byte) 0xF0, (byte) 0x68, (byte) 0x9D, 
		(byte) 0x08, (byte) 0xF0, (byte) 0x60, (byte) 0x4A, (byte) 0x6A, (byte) 0x6A, (byte) 0x6D, (byte) 0x5B, 
		(byte) 0xF1, (byte) 0x8D, (byte) 0x2D, (byte) 0xF0, (byte) 0x68, (byte) 0x6D, (byte) 0x5C, (byte) 0xF1, 
		(byte) 0x8D, (byte) 0x2E, (byte) 0xF0, (byte) 0x60, (byte) 0x4A, (byte) 0x90, (byte) 0x03, (byte) 0x4C, 
		(byte) 0xD3, (byte) 0xF9, (byte) 0x4A, (byte) 0xB0, (byte) 0x40, (byte) 0x4A, (byte) 0xB0, (byte) 0x17, 
		(byte) 0x4A, (byte) 0xB0, (byte) 0x0F, (byte) 0x68, (byte) 0x8D, (byte) 0x27, (byte) 0xF0, (byte) 0x4A, 
		(byte) 0x4A, (byte) 0x4A, (byte) 0xA8, (byte) 0xB9, (byte) 0xAF, (byte) 0xF1, (byte) 0x8D, (byte) 0x28, 
		(byte) 0xF0, (byte) 0x60, (byte) 0x68, (byte) 0x9D, (byte) 0x5D, (byte) 0xF0, (byte) 0x60, (byte) 0x4A, 
		(byte) 0xB0, (byte) 0x05, (byte) 0x68, (byte) 0x8D, (byte) 0x01, (byte) 0xF0, (byte) 0x60, (byte) 0x68, 
		(byte) 0xF0, (byte) 0x11, (byte) 0x9D, (byte) 0x75, (byte) 0xF0, (byte) 0xBC, (byte) 0x63, (byte) 0xF0, 
		(byte) 0xD0, (byte) 0x08, (byte) 0x9D, (byte) 0x63, (byte) 0xF0, (byte) 0xA9, (byte) 0x01, (byte) 0x9D, 
		(byte) 0x66, (byte) 0xF0, (byte) 0x60, (byte) 0x9D, (byte) 0x63, (byte) 0xF0, (byte) 0x9D, (byte) 0x69, 
		(byte) 0xF0, (byte) 0x9D, (byte) 0x6C, (byte) 0xF0, (byte) 0x60, (byte) 0x4A, (byte) 0xB0, (byte) 0x30, 
		(byte) 0x4A, (byte) 0xB0, (byte) 0x05, (byte) 0x68, (byte) 0x9D, (byte) 0x39, (byte) 0xF0, (byte) 0x60, 
		(byte) 0x68, (byte) 0xA0, (byte) 0x00, (byte) 0x4A, (byte) 0x90, (byte) 0x02, (byte) 0xC8, (byte) 0x18, 
		(byte) 0x48, (byte) 0x29, (byte) 0x07, (byte) 0x79, (byte) 0xAC, (byte) 0xF1, (byte) 0x9D, (byte) 0x78, 
		(byte) 0xF0, (byte) 0x9D, (byte) 0x81, (byte) 0xF0, (byte) 0x68, (byte) 0x4A, (byte) 0x4A, (byte) 0x4A, 
		(byte) 0x18, (byte) 0x79, (byte) 0xAD, (byte) 0xF1, (byte) 0x9D, (byte) 0x7B, (byte) 0xF0, (byte) 0x9D, 
		(byte) 0x84, (byte) 0xF0, (byte) 0xA9, (byte) 0x5B, (byte) 0x9D, (byte) 0x7E, (byte) 0xF0, (byte) 0x60, 
		(byte) 0x4A, (byte) 0xB0, (byte) 0x05, (byte) 0x68, (byte) 0x9D, (byte) 0xA2, (byte) 0xF0, (byte) 0x60, 
		(byte) 0x68, (byte) 0x8D, (byte) 0xCC, (byte) 0xF0, (byte) 0x60, (byte) 0x4A, (byte) 0xB0, (byte) 0x27, 
		(byte) 0x4A, (byte) 0xB0, (byte) 0x0D, (byte) 0x4A, (byte) 0xB0, (byte) 0x05, (byte) 0x68, (byte) 0x8D, 
		(byte) 0x29, (byte) 0xF0, (byte) 0x60, (byte) 0x68, (byte) 0x9D, (byte) 0x9F, (byte) 0xF0, (byte) 0x60, 
		(byte) 0x4A, (byte) 0xB0, (byte) 0x0F, (byte) 0x68, (byte) 0x9D, (byte) 0x93, (byte) 0xF0, (byte) 0xA0, 
		(byte) 0x00, (byte) 0x0A, (byte) 0x90, (byte) 0x01, (byte) 0x88, (byte) 0x98, (byte) 0x9D, (byte) 0x96, 
		(byte) 0xF0, (byte) 0x60, (byte) 0x68, (byte) 0x9D, (byte) 0x72, (byte) 0xF0, (byte) 0x60, (byte) 0x4A, 
		(byte) 0xB0, (byte) 0x1C, (byte) 0x4A, (byte) 0xB0, (byte) 0x15, (byte) 0x68, (byte) 0x9D, (byte) 0xB7, 
		(byte) 0xF0, (byte) 0xA5, (byte) 0xFD, (byte) 0x9D, (byte) 0xB1, (byte) 0xF0, (byte) 0xA5, (byte) 0xFE, 
		(byte) 0x9D, (byte) 0xB4, (byte) 0xF0, (byte) 0xBD, (byte) 0x33, (byte) 0xF0, (byte) 0x9D, (byte) 0xAE, 
		(byte) 0xF0, (byte) 0x60, (byte) 0x68, (byte) 0x6C, (byte) 0x5F, (byte) 0xF1, (byte) 0x4A, (byte) 0xB0, 
		(byte) 0x1E, (byte) 0x68, (byte) 0xD0, (byte) 0x0A, (byte) 0x9D, (byte) 0x4B, (byte) 0xF0, (byte) 0x9D, 
		(byte) 0x51, (byte) 0xF0, (byte) 0x9D, (byte) 0x54, (byte) 0xF0, (byte) 0x60, (byte) 0x9D, (byte) 0x5A, 
		(byte) 0xF0, (byte) 0xBC, (byte) 0x4B, (byte) 0xF0, (byte) 0xD0, (byte) 0x08, (byte) 0x9D, (byte) 0x4B, 
		(byte) 0xF0, (byte) 0xA9, (byte) 0x01, (byte) 0x9D, (byte) 0x4E, (byte) 0xF0, (byte) 0x60, (byte) 0x68, 
		(byte) 0x9D, (byte) 0x2A, (byte) 0xF0, (byte) 0x60, (byte) 0x4A, (byte) 0x90, (byte) 0x08, (byte) 0x9D, 
		(byte) 0x48, (byte) 0xF0, (byte) 0x68, (byte) 0x9D, (byte) 0x45, (byte) 0xF0, (byte) 0x60, (byte) 0x68, 
		(byte) 0x4A, (byte) 0xB0, (byte) 0x61, (byte) 0x4A, (byte) 0xB0, (byte) 0x25, (byte) 0x4A, (byte) 0xB0, 
		(byte) 0x05, (byte) 0x4A, (byte) 0xA0, (byte) 0xF0, (byte) 0xD0, (byte) 0x06, (byte) 0x0A, (byte) 0x0A, 
		(byte) 0x0A, (byte) 0x0A, (byte) 0xA0, (byte) 0x0F, (byte) 0x85, (byte) 0xFF, (byte) 0x98, (byte) 0xB0, 
		(byte) 0x09, (byte) 0x3D, (byte) 0x1D, (byte) 0xF0, (byte) 0x05, (byte) 0xFF, (byte) 0x9D, (byte) 0x1D, 
		(byte) 0xF0, (byte) 0x60, (byte) 0x3D, (byte) 0x20, (byte) 0xF0, (byte) 0x05, (byte) 0xFF, (byte) 0x9D, 
		(byte) 0x20, (byte) 0xF0, (byte) 0x60, (byte) 0x4A, (byte) 0xB0, (byte) 0x38, (byte) 0x4A, (byte) 0xB0, 
		(byte) 0x64, (byte) 0x85, (byte) 0xFF, (byte) 0xBD, (byte) 0xBA, (byte) 0xF0, (byte) 0xDD, (byte) 0xA9, 
		(byte) 0xF1, (byte) 0xF0, (byte) 0x54, (byte) 0xFE, (byte) 0xBA, (byte) 0xF0, (byte) 0xA8, (byte) 0xA5, 
		(byte) 0xFD, (byte) 0x99, (byte) 0xE1, (byte) 0xF0, (byte) 0xA5, (byte) 0xFE, (byte) 0x99, (byte) 0xF0, 
		(byte) 0xF0, (byte) 0xBD, (byte) 0x33, (byte) 0xF0, (byte) 0x99, (byte) 0x2F, (byte) 0xF1, (byte) 0xA4, 
		(byte) 0xFF, (byte) 0xB9, (byte) 0x17, (byte) 0xF1, (byte) 0xF0, (byte) 0x36, (byte) 0x85, (byte) 0xFE, 
		(byte) 0xB9, (byte) 0xFF, (byte) 0xF0, (byte) 0x85, (byte) 0xFD, (byte) 0xB9, (byte) 0x3E, (byte) 0xF1, 
		(byte) 0x9D, (byte) 0x33, (byte) 0xF0, (byte) 0x60, (byte) 0xB0, (byte) 0x4B, (byte) 0x4A, (byte) 0xB0, 
		(byte) 0x3C, (byte) 0xA8, (byte) 0xA5, (byte) 0xFD, (byte) 0x99, (byte) 0xFF, (byte) 0xF0, (byte) 0xA5, 
		(byte) 0xFE, (byte) 0x99, (byte) 0x17, (byte) 0xF1, (byte) 0xBD, (byte) 0x33, (byte) 0xF0, (byte) 0x99, 
		(byte) 0x3E, (byte) 0xF1, (byte) 0xBD, (byte) 0xBA, (byte) 0xF0, (byte) 0xDD, (byte) 0xA9, (byte) 0xF1, 
		(byte) 0xF0, (byte) 0x0D, (byte) 0xFE, (byte) 0xBA, (byte) 0xF0, (byte) 0xA8, (byte) 0xA9, (byte) 0x00, 
		(byte) 0x99, (byte) 0xF0, (byte) 0xF0, (byte) 0x60, (byte) 0xA9, (byte) 0x30, (byte) 0x2C, (byte) 0xA9, 
		(byte) 0x28, (byte) 0x8D, (byte) 0x00, (byte) 0xF0, (byte) 0x60, (byte) 0x0A, (byte) 0x0A, (byte) 0x0A, 
		(byte) 0x0A, (byte) 0x4D, (byte) 0x25, (byte) 0xF0, (byte) 0x29, (byte) 0xF0, (byte) 0x4D, (byte) 0x25, 
		(byte) 0xF0, (byte) 0x8D, (byte) 0x25, (byte) 0xF0, (byte) 0x60, (byte) 0x4D, (byte) 0x26, (byte) 0xF0, 
		(byte) 0x29, (byte) 0x0F, (byte) 0x4D, (byte) 0x26, (byte) 0xF0, (byte) 0x8D, (byte) 0x26, (byte) 0xF0, 
		(byte) 0x60, (byte) 0x4A, (byte) 0xB0, (byte) 0x0B, (byte) 0x4A, (byte) 0xB0, (byte) 0x04, (byte) 0x8D, 
		(byte) 0xCA, (byte) 0xF0, (byte) 0x60, (byte) 0x8D, (byte) 0xCB, (byte) 0xF0, (byte) 0x60, (byte) 0x4A, 
		(byte) 0x90, (byte) 0x03, (byte) 0x4C, (byte) 0xA5, (byte) 0xFB, (byte) 0x4A, (byte) 0xA8, (byte) 0xF0, 
		(byte) 0x21, (byte) 0x88, (byte) 0xF0, (byte) 0x34, (byte) 0x88, (byte) 0xF0, (byte) 0x42, (byte) 0x88, 
		(byte) 0xF0, (byte) 0x4A, (byte) 0x88, (byte) 0xF0, (byte) 0x52, (byte) 0x88, (byte) 0xF0, (byte) 0x5C, 
		(byte) 0x88, (byte) 0xF0, (byte) 0x66, (byte) 0x88, (byte) 0xF0, (byte) 0x73, (byte) 0x29, (byte) 0x07, 
		(byte) 0x09, (byte) 0x10, (byte) 0xB0, (byte) 0x03, (byte) 0x4C, (byte) 0xB7, (byte) 0xFA, (byte) 0x4C, 
		(byte) 0x7F, (byte) 0xFA, (byte) 0xAC, (byte) 0x26, (byte) 0xF0, (byte) 0xB0, (byte) 0x07, (byte) 0xC8, 
		(byte) 0x98, (byte) 0x29, (byte) 0x0F, (byte) 0xD0, (byte) 0x07, (byte) 0x60, (byte) 0x98, (byte) 0x29, 
		(byte) 0x0F, (byte) 0xF0, (byte) 0x04, (byte) 0x88, (byte) 0x8C, (byte) 0x26, (byte) 0xF0, (byte) 0x60, 
		(byte) 0xBD, (byte) 0x62, (byte) 0xF1, (byte) 0x49, (byte) 0xFF, (byte) 0x2D, (byte) 0x25, (byte) 0xF0, 
		(byte) 0x90, (byte) 0x03, (byte) 0x1D, (byte) 0x62, (byte) 0xF1, (byte) 0x8D, (byte) 0x25, (byte) 0xF0, 
		(byte) 0x60, (byte) 0xBD, (byte) 0x1A, (byte) 0xF0, (byte) 0x29, (byte) 0xFB, (byte) 0x90, (byte) 0x55, 
		(byte) 0x09, (byte) 0x04, (byte) 0xB0, (byte) 0x51, (byte) 0xBD, (byte) 0x1A, (byte) 0xF0, (byte) 0x29, 
		(byte) 0xFD, (byte) 0x90, (byte) 0x4A, (byte) 0x09, (byte) 0x02, (byte) 0xB0, (byte) 0x46, (byte) 0xAD, 
		(byte) 0x25, (byte) 0xF0, (byte) 0x29, (byte) 0xF7, (byte) 0x90, (byte) 0x02, (byte) 0x09, (byte) 0x08, 
		(byte) 0x8D, (byte) 0x25, (byte) 0xF0, (byte) 0x60, (byte) 0xAD, (byte) 0x26, (byte) 0xF0, (byte) 0x29, 
		(byte) 0x7F, (byte) 0x90, (byte) 0x02, (byte) 0x09, (byte) 0x80, (byte) 0x8D, (byte) 0x26, (byte) 0xF0, 
		(byte) 0x60, (byte) 0x98, (byte) 0x8D, (byte) 0xBD, (byte) 0xF0, (byte) 0x8D, (byte) 0xDF, (byte) 0xF0, 
		(byte) 0xC8, (byte) 0x8C, (byte) 0xE0, (byte) 0xF0, (byte) 0x2A, (byte) 0x8D, (byte) 0xC9, (byte) 0xF0, 
		(byte) 0x60, (byte) 0x98, (byte) 0x2A, (byte) 0x9D, (byte) 0x60, (byte) 0xF0, (byte) 0x60, (byte) 0x4A, 
		(byte) 0xB0, (byte) 0x27, (byte) 0x4A, (byte) 0xB0, (byte) 0x14, (byte) 0xD0, (byte) 0x02, (byte) 0xA9, 
		(byte) 0x08, (byte) 0x0A, (byte) 0x0A, (byte) 0x0A, (byte) 0x0A, (byte) 0x5D, (byte) 0x1A, (byte) 0xF0, 
		(byte) 0x29, (byte) 0xF0, (byte) 0x5D, (byte) 0x1A, (byte) 0xF0, (byte) 0x9D, (byte) 0x1A, (byte) 0xF0, 
		(byte) 0x60, (byte) 0x0A, (byte) 0x0A, (byte) 0x0A, (byte) 0x0A, (byte) 0x4D, (byte) 0x26, (byte) 0xF0, 
		(byte) 0x29, (byte) 0x70, (byte) 0x4D, (byte) 0x26, (byte) 0xF0, (byte) 0x8D, (byte) 0x26, (byte) 0xF0, 
		(byte) 0x60, (byte) 0x4A, (byte) 0x90, (byte) 0x04, (byte) 0x9D, (byte) 0xC0, (byte) 0xF0, (byte) 0x60, 
		(byte) 0xA8, (byte) 0xF0, (byte) 0x20, (byte) 0x88, (byte) 0xF0, (byte) 0x40, (byte) 0x88, (byte) 0xF0, 
		(byte) 0x63, (byte) 0x29, (byte) 0x03, (byte) 0x9D, (byte) 0xC3, (byte) 0xF0, (byte) 0xA9, (byte) 0x00, 
		(byte) 0x9D, (byte) 0xCD, (byte) 0xF0, (byte) 0x9D, (byte) 0xD0, (byte) 0xF0, (byte) 0x9D, (byte) 0xD3, 
		(byte) 0xF0, (byte) 0x9D, (byte) 0xD6, (byte) 0xF0, (byte) 0x8D, (byte) 0xD9, (byte) 0xF0, (byte) 0x8D, 
		(byte) 0xDC, (byte) 0xF0, (byte) 0x60, (byte) 0xBD, (byte) 0xB7, (byte) 0xF0, (byte) 0xF0, (byte) 0x05, 
		(byte) 0xDE, (byte) 0xB7, (byte) 0xF0, (byte) 0xF0, (byte) 0x12, (byte) 0xBD, (byte) 0x33, (byte) 0xF0, 
		(byte) 0xDD, (byte) 0xAE, (byte) 0xF0, (byte) 0xD0, (byte) 0x0B, (byte) 0xBD, (byte) 0xB1, (byte) 0xF0, 
		(byte) 0x85, (byte) 0xFD, (byte) 0xBD, (byte) 0xB4, (byte) 0xF0, (byte) 0x85, (byte) 0xFE, (byte) 0x60, 
		(byte) 0xA9, (byte) 0x38, (byte) 0x8D, (byte) 0x00, (byte) 0xF0, (byte) 0x60, (byte) 0xBD, (byte) 0xBA, 
		(byte) 0xF0, (byte) 0xDD, (byte) 0xA8, (byte) 0xF1, (byte) 0xF0, (byte) 0x18, (byte) 0xDE, (byte) 0xBA, 
		(byte) 0xF0, (byte) 0xA8, (byte) 0x88, (byte) 0xB9, (byte) 0xF0, (byte) 0xF0, (byte) 0xF0, (byte) 0x0D, 
		(byte) 0x85, (byte) 0xFE, (byte) 0xB9, (byte) 0xE1, (byte) 0xF0, (byte) 0x85, (byte) 0xFD, (byte) 0xB9, 
		(byte) 0x2F, (byte) 0xF1, (byte) 0x9D, (byte) 0x33, (byte) 0xF0, (byte) 0x60, (byte) 0xA9, (byte) 0x20, 
		(byte) 0x8D, (byte) 0x00, (byte) 0xF0, (byte) 0x60, (byte) 0xAD, (byte) 0x00, (byte) 0xF0, (byte) 0x5D, 
		(byte) 0x62, (byte) 0xF1, (byte) 0x8D, (byte) 0x00, (byte) 0xF0, (byte) 0xA9, (byte) 0x01, (byte) 0x9D, 
		(byte) 0x30, (byte) 0xF0, (byte) 0x60, (byte) 0xAD, (byte) 0x00, (byte) 0xF0, (byte) 0x29, (byte) 0x07, 
		(byte) 0x8D, (byte) 0x81, (byte) 0xFC, (byte) 0xD0, (byte) 0x03, (byte) 0x20, (byte) 0xE9, (byte) 0xF2, 
		(byte) 0x60, (byte) 0x00, (byte) 0xA2, (byte) 0x51, (byte) 0xA0, (byte) 0xFC, (byte) 0x8E, (byte) 0x5D, 
		(byte) 0xF1, (byte) 0x8C, (byte) 0x5E, (byte) 0xF1, (byte) 0x20, (byte) 0xCF, (byte) 0xF1, (byte) 0xA2, 
		(byte) 0x00, (byte) 0xA0, (byte) 0x09, (byte) 0x20, (byte) 0x00, (byte) 0xF2, (byte) 0xA9, (byte) 0x07, 
		(byte) 0x8D, (byte) 0x00, (byte) 0xF0, (byte) 0x8D, (byte) 0x81, (byte) 0xFC, (byte) 0x60, (byte) 0x00, 
		(byte) 0x00, (byte) 0x00, (byte) 0xA9, (byte) 0x00, (byte) 0x29, (byte) 0xFF, (byte) 0xF0, (byte) 0xF6, 
		(byte) 0x4C, (byte) 0x29, (byte) 0xF3, (byte) 0xA9, (byte) 0x07, (byte) 0x8D, (byte) 0x00, (byte) 0xF0, 
		(byte) 0x60, (byte) 0x00, (byte) 0x20, (byte) 0x60, (byte) 0xEC, (byte) 0x4C, (byte) 0x60, (byte) 0xFC, 
		(byte) 0x20, (byte) 0x80, (byte) 0xEC, (byte) 0x4C, (byte) 0x80, (byte) 0xFC, 
	};

	private void installPlayer(final byte[] c64buf) {
		int dest = (_sidtune_sidplayer1[0] & 0xff) + ((_sidtune_sidplayer1[1] & 0xff) << 8);
		System.arraycopy(_sidtune_sidplayer1, 2+0, c64buf, dest, _sidtune_sidplayer1.length - 2);
		// Point player #1 to data #1.
		c64buf[dest + 0xc6e] = MUS_DATA_ADDR + 2 & 0xFF;
		c64buf[dest + 0xc70] = MUS_DATA_ADDR + 2 >> 8;

		if (info.sidChipBase2 != 0) {
			// Install MUS player #2.
			dest = (_sidtune_sidplayer2[0] & 0xff) + ((_sidtune_sidplayer2[1] & 0xff) << 8);
			System.arraycopy(_sidtune_sidplayer2, 2, c64buf, dest,
					_sidtune_sidplayer2.length - 2);
			// Point player #2 to data #2.
			c64buf[dest + 0xc6e] = (byte) (MUS_DATA_ADDR
					+ musDataLen + 2 & 0xFF);
			c64buf[dest + 0xc70] = (byte) (MUS_DATA_ADDR
					+ musDataLen + 2 >> 8);
		}
	}
}
